# Copyright (c) 2025 Contributors to COVESA
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License 2.0 which is available at
# https://www.mozilla.org/en-US/MPL/2.0/
#
# SPDX-License-Identifier: MPL-2.0

from pathlib import Path

from graphql import build_schema, print_schema
from vss_tools.exporters.s2dm import (
    generate_s2dm_schema,
    get_metadata_df,
    print_schema_with_vspec_directives,
)
from vss_tools.exporters.s2dm.exporter import S2DM_CONVERSIONS
from vss_tools.main import get_trees
from vss_tools.utils.graphql_utils import GraphQLElementType, convert_name_for_graphql_schema


class TestS2DMExporter:
    """Test class for S2DM GraphQL exporter."""

    def test_convert_for_graphql_type_names(self):
        """Test conversion to GraphQL type names using unified function."""
        assert (
            convert_name_for_graphql_schema("Vehicle.Cabin.Seat", GraphQLElementType.TYPE, S2DM_CONVERSIONS)
            == "Vehicle_Cabin_Seat"
        )
        assert convert_name_for_graphql_schema("simple", GraphQLElementType.TYPE, S2DM_CONVERSIONS) == "Simple"
        assert convert_name_for_graphql_schema("with-dash", GraphQLElementType.TYPE, S2DM_CONVERSIONS) == "WithDash"

    def test_convert_for_graphql_field_names(self):
        """Test conversion to GraphQL field names using unified function."""
        assert convert_name_for_graphql_schema("SomeName", GraphQLElementType.FIELD, S2DM_CONVERSIONS) == "someName"
        assert convert_name_for_graphql_schema("simple", GraphQLElementType.FIELD, S2DM_CONVERSIONS) == "simple"
        assert convert_name_for_graphql_schema("with.dots", GraphQLElementType.FIELD, S2DM_CONVERSIONS) == "withDots"

    def test_get_metadata_df_with_seat_example(self):
        """Test metadata extraction from seat example."""
        # Load the example seat vspec
        tree, _ = get_trees(
            vspec=Path("tests/vspec/test_s2dm/example_seat.vspec"),
            include_dirs=(),
            aborts=(),
            strict=False,
            extended_attributes=(),
            quantities=(Path("tests/vspec/test_s2dm/test_quantities.yaml"),),
            units=(Path("tests/vspec/test_s2dm/test_units.yaml"),),
            overlays=(),
            expand=False,
        )

        branches_df, leaves_df = get_metadata_df(tree)

        # Check that we have expected branches
        assert "Vehicle" in branches_df.index
        assert "Vehicle.Cabin" in branches_df.index
        assert "Vehicle.Cabin.Seat" in branches_df.index

        # Check that we have expected leaves
        assert "Vehicle.Cabin.DriverPosition" in leaves_df.index
        assert "Vehicle.Cabin.Seat.HeatingCooling" in leaves_df.index

        # Check instance information
        seat_row = branches_df.loc["Vehicle.Cabin.Seat"]
        assert seat_row["instances"] is not None and seat_row["instances"] != ""

    def test_generate_s2dm_schema_basic_structure(self):
        """Test that basic schema generation works."""
        # Load the example seat vspec
        tree, _ = get_trees(
            vspec=Path("tests/vspec/test_s2dm/example_seat.vspec"),
            include_dirs=(),
            aborts=(),
            strict=False,
            extended_attributes=(),
            quantities=(Path("tests/vspec/test_s2dm/test_quantities.yaml"),),
            units=(Path("tests/vspec/test_s2dm/test_units.yaml"),),
            overlays=(),
            expand=False,
        )

        schema, _, _, _ = generate_s2dm_schema(tree)

        # Check that schema is valid
        assert schema is not None

        # Check that Query type exists
        query_type = schema.type_map["Query"]
        assert query_type is not None

        # Check that Vehicle type exists
        vehicle_type = schema.type_map["Vehicle"]
        assert vehicle_type is not None

        # Check that custom scalars exist
        assert "Int8" in schema.type_map
        assert "UInt8" in schema.type_map
        assert "UInt16" in schema.type_map

        # Check that custom directives exist
        directive_names = [d.name for d in schema.directives]
        assert "vspec" in directive_names
        assert "range" in directive_names
        assert "instanceTag" in directive_names

    def test_schema_can_be_printed(self):
        """Test that the schema can be serialized to GraphQL SDL."""
        # Load the example seat vspec
        tree, _ = get_trees(
            vspec=Path("tests/vspec/test_s2dm/example_seat.vspec"),
            include_dirs=(),
            aborts=(),
            strict=False,
            extended_attributes=(),
            quantities=(Path("tests/vspec/test_s2dm/test_quantities.yaml"),),
            units=(Path("tests/vspec/test_s2dm/test_units.yaml"),),
            overlays=(),
            expand=False,
        )

        schema, _, _, _ = generate_s2dm_schema(tree)
        schema_str = print_schema(schema)

        # Check that output contains expected elements
        assert "type Query" in schema_str
        assert "type Vehicle" in schema_str
        assert "scalar Int8" in schema_str
        assert "directive @vspec" in schema_str

        # Check that it can be parsed back
        parsed_schema = build_schema(schema_str)
        assert parsed_schema is not None

    def test_unit_enums_generation(self):
        """Test that unit enums are generated correctly."""
        # Load the example seat vspec
        tree, _ = get_trees(
            vspec=Path("tests/vspec/test_s2dm/example_seat.vspec"),
            include_dirs=(),
            aborts=(),
            strict=False,
            extended_attributes=(),
            quantities=(Path("tests/vspec/test_s2dm/test_quantities.yaml"),),
            units=(Path("tests/vspec/test_s2dm/test_units.yaml"),),
            overlays=(),
            expand=False,
        )

        schema, _, _, _ = generate_s2dm_schema(tree)
        schema_str = print_schema(schema)

        # Check that unit enums are generated
        assert "enum LengthUnitEnum" in schema_str
        assert "enum AngleUnitEnum" in schema_str
        assert "enum RelationUnitEnum" in schema_str

        # Check that enum values use uppercase unit names
        assert "MILLIMETER" in schema_str
        assert "DEGREE" in schema_str
        assert "PERCENT" in schema_str

        # Check that unit arguments are added to fields with proper defaults
        assert "unit: LengthUnitEnum = MILLIMETER" in schema_str
        assert "unit: AngleUnitEnum = DEGREE" in schema_str
        assert "unit: RelationUnitEnum = PERCENT" in schema_str

    def test_vspec_comment_directives(self):
        """Test that @vspec comment directives are generated correctly."""
        # Load the example seat vspec
        tree, _ = get_trees(
            vspec=Path("tests/vspec/test_s2dm/example_seat.vspec"),
            include_dirs=(),
            aborts=(),
            strict=False,
            extended_attributes=(),
            quantities=(Path("tests/vspec/test_s2dm/test_quantities.yaml"),),
            units=(Path("tests/vspec/test_s2dm/test_units.yaml"),),
            overlays=(),
            expand=False,
        )

        schema, unit_enums_metadata, allowed_enums_metadata, vspec_comments = generate_s2dm_schema(tree)
        schema_str = print_schema_with_vspec_directives(
            schema, unit_enums_metadata, allowed_enums_metadata, vspec_comments
        )

        # Check that field comments are captured
        assert len(vspec_comments["field_comments"]) > 0

        # Check that type comments are captured
        assert len(vspec_comments["type_comments"]) > 0

        # Check that @vspec comment directives appear in the output with new consolidated format
        assert 'comment: "Affects the property (SingleSeat.Position)."' in schema_str

        # Check for specific field comment directive
        assert "Affects the property (SingleSeat.Position)" in schema_str

        # Check for type comment directive
        assert "Seating is here considered as the part of the seat that supports the thighs" in schema_str

    def test_range_and_deprecation_directives(self):
        """Test that @range and @deprecated directives are correctly generated for VSS constraints"""
        tree, _ = get_trees(
            vspec=Path("tests/vspec/test_s2dm/example_seat.vspec"),
            include_dirs=(),
            aborts=(),
            strict=False,
            extended_attributes=(),
            quantities=(Path("tests/vspec/test_s2dm/test_quantities.yaml"),),
            units=(Path("tests/vspec/test_s2dm/test_units.yaml"),),
            overlays=(),
            expand=False,
        )

        schema, unit_enums_metadata, allowed_enums_metadata, vspec_comments = generate_s2dm_schema(tree)
        sdl = print_schema_with_vspec_directives(schema, unit_enums_metadata, allowed_enums_metadata, vspec_comments)

        # Test @deprecated directive for massage field
        assert '@deprecated(reason: "v5.0 - refactored to Seat.MassageLevel")' in sdl

        # Test @range directives for fields with constraints
        assert "@range(min: -100, max: 100)" in sdl  # heatingCooling
        assert "@range(min: 0, max: 100)" in sdl  # massage, massageLevel, lumbarSupport, sideBolsterSupport

        # Test @range directives with only min constraint (e.g., height: min 0)
        assert "@range(min: 0)" in sdl

        # Ensure no empty @range directives are present
        assert "@range(min: , max: )" not in sdl
        assert "@range(min: , max:" not in sdl
        assert "@range(min:, max:" not in sdl

        # Check specific field combinations
        # massage field should have @vspec with VSS type, @deprecated and @range
        import re

        massage_field_pattern = (
            r"massage\([^)]*\):[^@]*@vspec\([^)]*vspecType: ACTUATOR[^)]*\)"
            r"[^@]*@deprecated[^@]*@range\(min: 0, max: 100\)"
        )
        assert re.search(massage_field_pattern, sdl) is not None

        # massageLevel should have @range but not @deprecated
        massageLevel_context = sdl[sdl.find("massageLevel") : sdl.find("massageLevel") + 200]
        assert "@range(min: 0, max: 100)" in massageLevel_context
        assert "@deprecated" not in massageLevel_context

    def test_instance_tag_support(self):
        """Test that @instanceTag directive and dimensional enums are correctly generated for VSS instances"""
        tree, _ = get_trees(
            vspec=Path("tests/vspec/test_s2dm/example_seat.vspec"),
            include_dirs=(),
            aborts=(),
            strict=False,
            extended_attributes=(),
            quantities=(Path("tests/vspec/test_s2dm/test_quantities.yaml"),),
            units=(Path("tests/vspec/test_s2dm/test_units.yaml"),),
            overlays=(),
            expand=False,
        )

        schema, unit_enums_metadata, allowed_enums_metadata, vspec_comments = generate_s2dm_schema(tree)
        sdl = print_schema_with_vspec_directives(schema, unit_enums_metadata, allowed_enums_metadata, vspec_comments)

        # Test that instance tag type is created
        assert "type Vehicle_Cabin_Seat_InstanceTag @instanceTag" in sdl

        # Test that dimensional enums are created with correct names and values
        assert "enum Vehicle_Cabin_Seat_InstanceTag_Dimension1" in sdl
        assert "enum Vehicle_Cabin_Seat_InstanceTag_Dimension2" in sdl

        # Test enum values match expected format from VSS instances
        assert "Row1" in sdl
        assert "Row2" in sdl
        assert "DriverSide" in sdl
        assert "PassengerSide" in sdl

        # Test that main type has instanceTag field
        assert "instanceTag: Vehicle_Cabin_Seat_InstanceTag" in sdl

        # Test that instance tag type has dimension fields
        assert "dimension1: Vehicle_Cabin_Seat_InstanceTag_Dimension1" in sdl
        assert "dimension2: Vehicle_Cabin_Seat_InstanceTag_Dimension2" in sdl

        # Test that main type does NOT have @instanceTag directive (only instance tag type has it)
        # Look specifically for the main type, not sub-types like Vehicle_Cabin_Seat_Airbag
        main_type_pattern = "type Vehicle_Cabin_Seat "
        main_type_lines = [
            line
            for line in sdl.split("\n")
            if main_type_pattern in line
            and "{" in line
            and "@instanceTag" not in line
            and "Vehicle_Cabin_Seat_" not in line.replace(main_type_pattern, "")
        ]
        assert len(main_type_lines) == 1
        assert "@instanceTag" not in main_type_lines[0]

        # Test that types with instances get ID field
        assert "id: ID!" in sdl

        # Verify the complete structure matches the reference pattern
        # The seat should be a list field (seat_s) because it has instances
        assert "seat_s: [Vehicle_Cabin_Seat]" in sdl

    def test_allowed_value_enums_generation(self):
        """Test that allowed value enums are generated correctly."""
        # Use S2DM-specific test file with allowed values
        test_vspec_path = Path(__file__).parent / "vspec" / "test_s2dm" / "test_allowed_values.vspec"

        # Get tree from file
        tree, _ = get_trees(
            vspec=test_vspec_path,
            include_dirs=(),
            aborts=(),
            strict=False,
            extended_attributes=(),
            quantities=(),
            units=(),
            overlays=(),
        )

        # Generate schema
        schema, _, _, _ = generate_s2dm_schema(tree)
        schema_sdl = print_schema(schema)

        # Check that allowed value enums were generated
        # Check for the string field with allowed values ["PARK", "REVERSE", "NEUTRAL", "DRIVE"]
        assert "Vehicle_Transmission_GearMode_Enum" in schema_sdl
        assert "PARK" in schema_sdl
        assert "REVERSE" in schema_sdl
        assert "NEUTRAL" in schema_sdl
        assert "DRIVE" in schema_sdl

        # Check for numeric field with allowed values [1, 2, 3, 4, 5]
        assert "Vehicle_Performance_Level_Enum" in schema_sdl
        assert "_1" in schema_sdl  # Numeric values should have underscore prefix
        assert "_2" in schema_sdl
        assert "_5" in schema_sdl

        # Check for float field with allowed values [1.0, 2.5, 4.0, 5.0]
        assert "Vehicle_Performance_Rating_Enum" in schema_sdl
        assert "_1" in schema_sdl  # 1.0 becomes _1
        assert "_2_DOT_5" in schema_sdl  # 2.5 should use _DOT_
        assert "_4" in schema_sdl
        assert "_5" in schema_sdl

        # Check that fields use the enum types instead of base types
        assert "gearMode: Vehicle_Transmission_GearMode_Enum" in schema_sdl
        assert "level: Vehicle_Performance_Level_Enum" in schema_sdl
        assert "rating: Vehicle_Performance_Rating_Enum" in schema_sdl

        # Check enum descriptions
        assert "Allowed values for Vehicle.Transmission.GearMode" in schema_sdl

    def test_modular_export_flat_domains(self, tmp_path):
        """Test modular export with flat domain structure."""
        from vss_tools.exporters.s2dm.exporter import write_modular_schema

        # Load the example seat vspec
        tree, _ = get_trees(
            vspec=Path("tests/vspec/test_s2dm/example_seat.vspec"),
            include_dirs=(),
            aborts=(),
            strict=False,
            extended_attributes=(),
            quantities=(Path("tests/vspec/test_s2dm/test_quantities.yaml"),),
            units=(Path("tests/vspec/test_s2dm/test_units.yaml"),),
            overlays=(),
            expand=False,
        )

        # Generate schema
        schema, unit_enums_metadata, allowed_enums_metadata, vspec_comments = generate_s2dm_schema(tree)

        # Test modular export with flat domains (default)
        output_dir = tmp_path / "modular_flat"
        write_modular_schema(
            schema, unit_enums_metadata, allowed_enums_metadata, vspec_comments, output_dir, flat_domains=True
        )

        # Check that common files were created in other/ directory
        assert (output_dir / "other" / "directives.graphql").exists()
        assert (output_dir / "other" / "scalars.graphql").exists()
        assert (output_dir / "other" / "queries.graphql").exists()

        # Check that domain files were created in domain/ directory (flat structure)
        assert (output_dir / "domain" / "Vehicle.graphql").exists()
        assert (output_dir / "domain" / "Vehicle_Cabin.graphql").exists()
        assert (output_dir / "domain" / "Vehicle_Cabin_Seat.graphql").exists()

        # Check that unit enum files were created in other/ directory
        assert (output_dir / "other" / "units.graphql").exists()

        # Check that instance files were created in instances/ directory
        assert (output_dir / "instances" / "Vehicle_Cabin_Seat_InstanceTag.graphql").exists()

        # Verify content of a domain file
        vehicle_content = (output_dir / "domain" / "Vehicle.graphql").read_text()
        assert "type Vehicle" in vehicle_content
        assert "Domain-specific GraphQL types" in vehicle_content

    def test_modular_export_nested_domains(self, tmp_path):
        """Test modular export with nested domain structure."""
        from vss_tools.exporters.s2dm.exporter import write_modular_schema

        # Load the example seat vspec
        tree, _ = get_trees(
            vspec=Path("tests/vspec/test_s2dm/example_seat.vspec"),
            include_dirs=(),
            aborts=(),
            strict=False,
            extended_attributes=(),
            quantities=(Path("tests/vspec/test_s2dm/test_quantities.yaml"),),
            units=(Path("tests/vspec/test_s2dm/test_units.yaml"),),
            overlays=(),
            expand=False,
        )

        # Generate schema
        schema, unit_enums_metadata, allowed_enums_metadata, vspec_comments = generate_s2dm_schema(tree)

        # Test modular export with nested domains
        output_dir = tmp_path / "modular_nested"
        write_modular_schema(
            schema, unit_enums_metadata, allowed_enums_metadata, vspec_comments, output_dir, flat_domains=False
        )

        # Check that common files were created in other/ directory
        assert (output_dir / "other" / "directives.graphql").exists()
        assert (output_dir / "other" / "queries.graphql").exists()

        # Check that nested structure was created in domain/ directory
        assert (output_dir / "domain" / "Vehicle.graphql").exists()
        assert (output_dir / "domain" / "Vehicle" / "Cabin.graphql").exists()
        assert (output_dir / "domain" / "Vehicle" / "Cabin" / "Seat.graphql").exists()

        # Check that enum directory structure was created in other/ directory
        assert (output_dir / "other" / "units.graphql").exists()

        # Check that instance files were created in instances/ directory
        assert (output_dir / "instances" / "Vehicle_Cabin_Seat_InstanceTag.graphql").exists()

        # Verify nested directory content
        seat_content = (output_dir / "domain" / "Vehicle" / "Cabin" / "Seat.graphql").read_text()
        assert "Vehicle_Cabin_Seat" in seat_content



# Copyright (c) 2024 Contributors to COVESA
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License 2.0 which is available at
# https://www.mozilla.org/en-US/MPL/2.0/
#
# SPDX-License-Identifier: MPL-2.0

"""
GraphQL directive processing using string templates.

This module provides a hybrid approach that uses GraphQL-core for schema
validation and string templates for directive injection. It replaces the
brittle string manipulation approach with a more maintainable template system.
"""

from string import Template

from graphql import GraphQLSchema, print_schema

from vss_tools.utils.graphql_utils import GraphQLElementType, convert_name_for_graphql_schema


class GraphQLDirectiveProcessor:
    """Processes GraphQL schema SDL to inject VSS-specific directives using templates."""

    # Template definitions for different directive types
    TEMPLATES = {
        "vspec_fqn": Template('@vspec(source: {kind: FQN, value: "$fqn"}, vspecType: $vspec_type)'),
        "vspec_unit": Template(
            '@vspec(source: {kind: UNIT, value: "$unit_key", '
            'note: "Taken and converted from full name <$unit_name>."})'
        ),
        "vspec_allowed": Template('@vspec(source: {kind: ALLOWED_VALUE, value: "$original_value"})'),
        "vspec_quantity": Template('@vspec(source: {kind: QUANTITY_KIND, value: "$quantity"})'),
        "vspec_comment": Template('@vspec(comment: "$comment")'),
        "vspec_consolidated": Template("@vspec($parts)"),
        "range": Template("@range($args)"),
        "deprecated": Template('@deprecated(reason: "$reason")'),
    }

    def process_schema(
        self, schema: GraphQLSchema, unit_enums_metadata: dict, allowed_enums_metadata: dict, vspec_comments: dict
    ) -> str:
        """
        Process schema SDL to inject directives using templates.

        Args:
            schema: GraphQL schema from graphql-core
            unit_enums_metadata: Unit enum metadata
            allowed_enums_metadata: Allowed enum metadata
            vspec_comments: VSS comment and directive data

        Returns:
            SDL string with injected directives
        """
        # Start with standard GraphQL-core SDL output
        sdl = print_schema(schema)
        lines = sdl.split("\n")

        # Process directives using templates
        processed_enum_values: set[str] = set()

        lines = self._process_unit_enum_directives(lines, unit_enums_metadata, processed_enum_values)
        lines = self._process_allowed_enum_directives(lines, allowed_enums_metadata, processed_enum_values)
        lines = self._process_field_directives(lines, vspec_comments)
        lines = self._process_deprecated_directives(lines, vspec_comments.get("field_deprecated", {}))
        lines = self._process_range_directives(lines, vspec_comments.get("field_ranges", {}))
        lines = self._process_type_directives(lines, vspec_comments)

        return "\n".join(lines)

    def _process_unit_enum_directives(
        self, lines: list[str], unit_enums_metadata: dict, processed_values: set
    ) -> list[str]:
        """Process unit enum directives using templates."""
        for quantity, units_data in unit_enums_metadata.items():
            enum_name = f"{convert_name_for_graphql_schema(quantity, GraphQLElementType.TYPE)}UnitEnum"

            in_target_enum = False
            for i, line in enumerate(lines):
                if line.strip().startswith(f"enum {enum_name}"):
                    if "@vspec" not in line:
                        directive = self.TEMPLATES["vspec_quantity"].substitute(quantity=quantity)
                        lines[i] = line.replace(" {", f" {directive} {{")
                    in_target_enum = True
                    continue
                elif line.strip().startswith("enum ") and in_target_enum:
                    in_target_enum = False
                    continue
                elif line.strip() == "}" and in_target_enum:
                    in_target_enum = False
                    continue

                if in_target_enum and line.strip() and not line.strip().startswith('"'):
                    stripped_line = line.strip()

                    for unit_key, unit_info in units_data.items():
                        unit_name = unit_info["name"]
                        enum_value_name = convert_name_for_graphql_schema(unit_name, GraphQLElementType.ENUM_VALUE)

                        enum_value_key = f"{enum_name}.{enum_value_name}"
                        if stripped_line.startswith(enum_value_name) and enum_value_key not in processed_values:
                            if "@vspec" not in line:
                                indent = line[: len(line) - len(line.lstrip())]
                                directive = self.TEMPLATES["vspec_unit"].substitute(
                                    unit_key=unit_key, unit_name=unit_name
                                )
                                lines[i] = f"{indent}{enum_value_name} {directive}"

                            processed_values.add(enum_value_key)
                            break
        return lines

    def _process_allowed_enum_directives(
        self, lines: list[str], allowed_enums_metadata: dict, processed_values: set
    ) -> list[str]:
        """Process allowed value enum directives using templates."""
        for enum_name, enum_data in allowed_enums_metadata.items():
            allowed_values_dict = enum_data.get("allowed_values", {})
            in_target_enum = False

            for i, line in enumerate(lines):
                if line.strip().startswith(f"enum {enum_name}"):
                    in_target_enum = True
                    continue
                elif line.strip().startswith("enum ") and in_target_enum:
                    in_target_enum = False
                    continue
                elif line.strip() == "}" and in_target_enum:
                    in_target_enum = False
                    continue

                if in_target_enum and line.strip() and not line.strip().startswith('"'):
                    stripped_line = line.strip()

                    for graphql_enum_value, original_value in allowed_values_dict.items():
                        enum_value_key = f"{enum_name}.{graphql_enum_value}"

                        if stripped_line.startswith(graphql_enum_value) and enum_value_key not in processed_values:
                            if "@vspec" not in line:
                                indent = line[: len(line) - len(line.lstrip())]
                                directive = self.TEMPLATES["vspec_allowed"].substitute(original_value=original_value)
                                lines[i] = f"{indent}{graphql_enum_value} {directive}"

                            processed_values.add(enum_value_key)
                            break
        return lines

    def _process_field_directives(self, lines: list[str], vspec_comments: dict) -> list[str]:
        """Process consolidated field @vspec directives using templates."""
        # Build consolidated field data
        field_data: dict[str, dict[str, str]] = {}

        # Merge VSS type information
        for field_path, vss_info in vspec_comments.get("field_vss_types", {}).items():
            if field_path not in field_data:
                field_data[field_path] = {}
            field_data[field_path]["fqn"] = vss_info["fqn"]
            field_data[field_path]["vspec_type"] = vss_info["vspec_type"]

        # Merge comments
        for field_path, comment in vspec_comments.get("field_comments", {}).items():
            if field_path not in field_data:
                field_data[field_path] = {}
            field_data[field_path]["comment"] = comment.replace('"', '\\"')

        # Apply consolidated directives
        for field_path, data in field_data.items():
            type_name, field_name = field_path.split(".")

            in_type = False
            for i, line in enumerate(lines):
                if line.strip().startswith(f"type {type_name}"):
                    in_type = True
                    continue
                elif line.strip().startswith("type ") and in_type:
                    in_type = False
                    continue
                elif line.strip() == "}" and in_type:
                    in_type = False
                    continue

                if in_type and line.strip().startswith(f"{field_name}") and "@vspec" not in line:
                    # Build directive parts
                    parts = []
                    if "fqn" in data:
                        parts.append(f'source: {{kind: FQN, value: "{data["fqn"]}"}}')
                    if "vspec_type" in data:
                        parts.append(f'vspecType: {data["vspec_type"]}')
                    if "comment" in data:
                        parts.append(f'comment: "{data["comment"]}"')

                    if parts:
                        directive = self.TEMPLATES["vspec_consolidated"].substitute(parts=", ".join(parts))
                        lines[i] = line.rstrip() + f" {directive}"

                    break

        return lines

    def _process_deprecated_directives(self, lines: list[str], field_deprecated: dict) -> list[str]:
        """Process @deprecated directives using templates."""
        for field_path, reason in field_deprecated.items():
            escaped_reason = reason.replace('"', '\\"')
            type_name, field_name = field_path.split(".")

            in_type = False
            for i, line in enumerate(lines):
                if line.strip().startswith(f"type {type_name}"):
                    in_type = True
                    continue
                elif line.strip().startswith("type ") and in_type:
                    in_type = False
                    continue
                elif line.strip() == "}" and in_type:
                    in_type = False
                    continue

                if (
                    in_type
                    and field_name in line
                    and ":" in line
                    and not line.strip().startswith('"')
                    and "@deprecated" not in line
                ):
                    insert_line = i + 1
                    while insert_line < len(lines) and lines[insert_line].strip().startswith("@vspec"):
                        insert_line += 1

                    if insert_line < len(lines) and "@deprecated" not in lines[insert_line]:
                        directive = self.TEMPLATES["deprecated"].substitute(reason=escaped_reason)
                        lines.insert(insert_line, f"    {directive}")
                    break
        return lines

    def _process_range_directives(self, lines: list[str], field_ranges: dict) -> list[str]:
        """Process @range directives using templates."""
        for field_path, range_data in field_ranges.items():
            range_args = []
            min_val = range_data.get("min")
            max_val = range_data.get("max")

            # Handle both None and empty string cases
            if min_val is not None and min_val != "":
                range_args.append(f"min: {min_val}")
            if max_val is not None and max_val != "":
                range_args.append(f"max: {max_val}")

            if not range_args:
                continue

            type_name, field_name = field_path.split(".")

            in_type = False
            for i, line in enumerate(lines):
                if line.strip().startswith(f"type {type_name}"):
                    in_type = True
                    continue
                elif line.strip().startswith("type ") and in_type:
                    in_type = False
                    continue
                elif line.strip() == "}" and in_type:
                    in_type = False
                    continue

                if (
                    in_type
                    and field_name in line
                    and ":" in line
                    and not line.strip().startswith('"')
                    and "@range" not in line
                ):
                    insert_line = i + 1
                    while insert_line < len(lines) and (
                        lines[insert_line].strip().startswith("@vspec")
                        or lines[insert_line].strip().startswith("@deprecated")
                    ):
                        insert_line += 1

                    if insert_line < len(lines) and "@range" not in lines[insert_line]:
                        directive = self.TEMPLATES["range"].substitute(args=", ".join(range_args))
                        lines.insert(insert_line, f"    {directive}")
                    break
        return lines

    def _process_type_directives(self, lines: list[str], vspec_comments: dict) -> list[str]:
        """Process type-level directives using templates."""
        for i, line in enumerate(lines):
            if line.strip().startswith("type "):
                type_line = line.strip()

                type_part = type_line[5:]  # Remove "type "
                if " {" in type_part:
                    type_name = type_part.split(" {")[0].strip()
                elif " @" in type_part:
                    type_name = type_part.split(" @")[0].strip()
                else:
                    continue

                needs_vspec_comment = type_name in vspec_comments.get("type_comments", {})
                needs_instance_tag = type_name in vspec_comments.get("instance_tags", {})
                needs_vss_type = type_name in vspec_comments.get("vss_types", {})

                if needs_vspec_comment or needs_instance_tag or needs_vss_type:
                    new_line = f"type {type_name}"

                    # Add @vspec directive with VSS type information
                    if needs_vss_type and "@vspec" not in line:
                        vss_info = vspec_comments["vss_types"][type_name]
                        fqn = vss_info["fqn"]
                        vspec_type = vss_info["vspec_type"]

                        parts = [f'source: {{kind: FQN, value: "{fqn}"}}, vspecType: {vspec_type}']

                        # Add comment if available
                        if needs_vspec_comment:
                            comment = vspec_comments["type_comments"][type_name]
                            escaped_comment = comment.replace('"', '\\"')
                            parts.append(f'comment: "{escaped_comment}"')

                        directive = self.TEMPLATES["vspec_consolidated"].substitute(parts=", ".join(parts))
                        new_line += f" {directive}"
                    elif needs_vspec_comment and "@vspec" not in line:
                        # Fallback to just comment
                        comment = vspec_comments["type_comments"][type_name]
                        escaped_comment = comment.replace('"', '\\"')
                        directive = self.TEMPLATES["vspec_comment"].substitute(comment=escaped_comment)
                        new_line += f" {directive}"

                    if needs_instance_tag and "@instanceTag" not in line:
                        new_line += " @instanceTag"

                    new_line += " {"
                    lines[i] = "  " + new_line  # Preserve original indentation

        return lines



# Copyright (c) 2022 Contributors to COVESA
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License 2.0 which is available at
# https://www.mozilla.org/en-US/MPL/2.0/
#
# SPDX-License-Identifier: MPL-2.0

from pathlib import Path

from vss_tools.main import get_trees

HERE = Path(__file__).resolve().parent


def test_exporters():
    tree, _ = get_trees(
        vspec=HERE / "test.vspec",
    )
    door = tree.children[0]
    row1 = door.children[0]
    assert row1.data.is_instance
    door.children[1]
    assert row1.data.is_instance
    special = row1.children[0]
    assert not special.data.is_instance



# Copyright (c) 2025 Contributors to COVESA, CEA LIST
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License 2.0 which is available at
# https://www.mozilla.org/en-US/MPL/2.0/
#
# SPDX-License-Identifier: MPL-2.0

from pathlib import Path

import rich_click as click

import vss_tools.cli_options as clo
from vss_tools import log
from vss_tools.main import get_trees
from vss_tools.model import VSSDataBranch
from vss_tools.tree import VSSNode

fqns: set[str] = set()


# make first character lowercase
def lc_first(s) -> str:
    if len(s) == 0:
        return s
    else:
        return s[0].lower() + s[1:]


# variant of get_fqn that skips instance nodes (and the top-level one)
def get_fqn2(node: VSSNode) -> str:
    fqn = node.name
    while node.parent:
        data = node.get_vss_data()
        if node.is_leaf or not (isinstance(data, VSSDataBranch) and data.is_instance):
            # generated classes are in a package carrying their own name, but enumerations are not
            if not getattr(data, "allowed", None):
                fqn = "P" + node.name + "." + fqn
        node = node.parent
    return fqn


def get_name(node: VSSNode, qualify: bool) -> str:
    if qualify:
        return get_fqn2(node)
    else:
        return node.name


# get the class name of a node that is not an instance going up.
# adds postfix IS
def get_classname(node: VSSNode, qualify: bool) -> str:
    data = node.get_vss_data()
    if node.is_leaf:
        return get_name(node, qualify)
    elif not (isinstance(data, VSSDataBranch) and data.is_instance):
        if has_nested_instance_child(node):
            return get_name(node, qualify) + "IS0"
        elif has_instance_child(node):
            return get_name(node, qualify) + "IS"
        else:
            return get_name(node, qualify)
    else:
        parent = node.parent
        if parent.get_vss_data().is_instance:
            parent = parent.parent
        if has_instance_child(node):
            # node is already instance, implies nested one
            return get_name(parent, qualify) + "_IS1"
        else:
            return get_name(parent, qualify)


# convenience function wrapping count_instance_children_depth
def has_instance_child(node: VSSNode) -> bool:
    return node.count_instance_children_depth() > 0


# convenience function wrapping count_instance_children_depth
def has_nested_instance_child(node: VSSNode) -> bool:
    return node.count_instance_children_depth() > 1


def get_enums(tree: VSSNode, fill: str, attributes: tuple[str]) -> str:
    tree_content_lines = []
    for node in tree.children:
        data = node.get_vss_data()
        if node.is_leaf:
            allowed = getattr(data, "allowed", None)
            fqn = get_fqn2(node)
            if allowed and (fqn not in fqns):
                # use enumeration instead of datatype, use 2nd level package name
                fqns.add(fqn)
                # create Enumeration
                tree_content_lines.append("")
                tree_content_lines.append("%s' %s" % (fill, data.description))
                tree_content_lines.append("%senum %s {" % (fill, node.name))
                for a in allowed:
                    tree_content_lines.append("%s\t%s," % (fill, a))
                tree_content_lines.append("%s}" % (fill))
        else:
            if not node.parent:
                # top level package, recurse only
                result = get_enums(node, fill + "\t", attributes)
                tree_content_lines.append(result)
            elif isinstance(data, VSSDataBranch) and data.is_instance:
                # instance node, recurse only (no package, no indent)
                result = get_enums(node, fill, attributes)
                if result:
                    tree_content_lines.append(result)
            else:
                result = get_enums(node, fill + "\t", attributes)
                # only add package, if it contains an enumeration
                if result:
                    tree_content_lines.append("")
                    tree_content_lines.append("%spackage P%s {" % (fill, node.name))
                    tree_content_lines.append(result)
                    tree_content_lines.append("%s}" % (fill))

    return "\n".join(tree_content_lines)


def get_rendered_class(tree: VSSNode, fill, attributes: tuple[str]) -> str:
    tree_content_lines = []
    for node in tree.children:
        data = node.get_vss_data()
        if node.is_leaf:
            # add an entry to the data type
            tree_content_lines.append("")
            tree_content_lines.append("%s' %s: %s" % (fill, data.type.value, data.description))
            datatype = getattr(data, "datatype", None)
            unit = getattr(data, "unit", None)
            if unit:
                tree_content_lines.append("%s' unit: %s" % (fill, unit))
            min = getattr(data, "min", None)
            max = getattr(data, "max", None)
            if (min is not None) and (max is not None):
                tree_content_lines.append("%s' interval [%s..%s]" % (fill, min, max))
            elif min is not None:
                tree_content_lines.append("%s' interval [%s..)" % (fill, min))
            elif max is not None:
                tree_content_lines.append("%s' interval (..%s]" % (fill, max))
            if getattr(data, "allowed", None):
                # use qualified name of enumeration as datatype
                datatype = get_fqn2(node)
            tree_content_lines.append("%s%s : %s" % (fill, lc_first(node.name), datatype))
        else:
            tree_content_lines.append("%s%s : %s" % (fill, lc_first(node.name), get_classname(node, True)))

    return "\n".join(tree_content_lines)


def get_rendered_tree(node: VSSNode, fill, attributes: tuple[str]) -> str:
    tree_content_lines = []
    data = node.get_vss_data()
    needPkg = node.parent and (node.is_leaf or not (isinstance(data, VSSDataBranch) and data.is_instance))
    if needPkg:
        tree_content_lines.append("%s' %s" % (fill, data.description))
        tree_content_lines.append("%spackage P%s {" % (fill, node.name))
        nFill = fill + "\t"
    else:
        nFill = fill
    tree_content_lines.append("%sclass %s {" % (nFill, get_classname(node, False)))
    # if the node has child instances, enter into first one (skip current child node)
    result = get_rendered_class(node, nFill + "\t", attributes)
    tree_content_lines.append(result)
    tree_content_lines.append("%s}" % (nFill))
    for node in node.children:
        data = node.get_vss_data()
        if node.is_leaf:
            pass
        else:
            result = get_rendered_tree(node, nFill, attributes)
            tree_content_lines.append(result)
            if isinstance(data, VSSDataBranch) and data.is_instance:
                break

    if needPkg:
        tree_content_lines.append("%s}" % (fill))

    return "\n".join(tree_content_lines)


@click.command()
@clo.vspec_opt
@clo.output_opt
@clo.include_dirs_opt
@clo.extended_attributes_opt
@clo.strict_opt
@clo.aborts_opt
@clo.expand_opt
@clo.overlays_opt
@clo.quantities_opt
@clo.units_opt
@clo.types_opt
@clo.strict_exceptions_opt
@click.option("--attr", help="Show VSSData attribute", multiple=True)
def cli(
    vspec: Path,
    include_dirs: tuple[Path],
    extended_attributes: tuple[str],
    strict: bool,
    aborts: tuple[str],
    expand: bool,
    overlays: tuple[Path],
    quantities: tuple[Path],
    units: tuple[Path],
    types: tuple[Path],
    output: Path | None,
    attr: tuple[str],
    strict_exceptions: Path | None,
):
    """
    Export as PlantUML.
    """
    tree, datatype_tree = get_trees(
        vspec=vspec,
        include_dirs=include_dirs,
        aborts=aborts,
        strict=strict,
        extended_attributes=extended_attributes,
        quantities=quantities,
        units=units,
        types=types,
        overlays=overlays,
        expand=expand,
        strict_exceptions_file=strict_exceptions,
    )

    plant_code = get_enums(tree, "", attr)
    if len(fqns) > 0:
        plant_code += "\n' --- end of enums\n\n"

    plant_code += get_rendered_tree(tree, "", attr) + "\n"
    if datatype_tree:
        plant_code += "\n'datatype tree:\n" + get_rendered_tree(datatype_tree, "", attr)

    if output:
        log.info(f"Writing tree to: {output.absolute()}")
        with open(output, "w") as f:
            f.write(plant_code)
    else:
        log.info(plant_code)



# Copyright (c) 2023 Contributors to COVESA
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License 2.0 which is available at
# https://www.mozilla.org/en-US/MPL/2.0/
#
# SPDX-License-Identifier: MPL-2.0

import subprocess
from pathlib import Path

import pytest

HERE = Path(__file__).resolve().parent
TEST_UNITS = HERE / ".." / "test_units.yaml"
TEST_QUANT = HERE / ".." / "test_quantities.yaml"


@pytest.mark.parametrize(
    "vspec_file, type_file, type_out_file",
    [
        ("no_description_branch.vspec", None, None),
        ("no_description_signal.vspec", None, None),
        ("correct.vspec", "no_description_type_branch.vspec", "ot.json"),
        ("correct.vspec", "no_description_type_struct.vspec", "ot.json"),
        ("correct.vspec", "no_description_type_property.vspec", "ot.json"),
    ],
)
def test_description_error(vspec_file: str, type_file: str, type_out_file: str, tmp_path):
    output = tmp_path / "out.json"
    log = tmp_path / "out.log"
    cmd = f"vspec --log-file {log} export json --pretty -u {TEST_UNITS} -q {TEST_QUANT}"
    if type_file:
        cmd += f" --types {HERE / type_file}"
    if type_out_file:
        cmd += f" --types-output {tmp_path / type_out_file}"
    cmd += f" --vspec {HERE / vspec_file} --output {output}"

    process = subprocess.run(cmd.split(), capture_output=True, text=True)
    assert process.returncode != 0
    log_content = log.read_text()
    print(log_content)
    assert "'type': 'assertion_error'" in log_content
    assert "1 model error(s):" in log_content



# Copyright (c) 2022 Contributors to COVESA
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License 2.0 which is available at
# https://www.mozilla.org/en-US/MPL/2.0/
#
# SPDX-License-Identifier: MPL-2.0

import re
from pathlib import Path

import vss_tools.model as model
from anytree import RenderTree
from vss_tools.main import get_trees
from vss_tools.tree import VSSNode

HERE = Path(__file__).resolve().parent
RESOURCES = HERE / "resources"


#  TEST SIMPLE INSTANCE STRUCTURE #
# Goal: Check if simple (=not nested) instantiation works
#
# How: Define three instances in different ways and check,
# if the result matches every time
#
# TEST SIMPLE INSTANCE STRUCTURE #


# Needed test files
TEST_FILES_SIMPLE = [
    "instance_simple_range.vspec",
    "instance_simple_enum.vspec",
    "instance_simple_range_list.vspec",
    "instance_simple_enum_list.vspec",
]


def test_simple_structures(request):
    for tfs in TEST_FILES_SIMPLE:
        # load the file
        print(tfs)
        tree, _ = get_trees(vspec=RESOURCES / tfs, include_dirs=(RESOURCES,))

        print(RenderTree(tree).by_attr())
        # check if root node has 3 children
        assert len(tree.children) == 3

        list_of_instances = ["Vehicle.Test1", "Vehicle.Test2", "Vehicle.Test3"]

        child: VSSNode
        for child in tree.children:
            assert child.get_fqn() in list_of_instances  # < list qualified names
            assert child.data.type == model.NodeType.BRANCH
            assert child.data.description == "High-level vehicle data."
            # < check that they have exactly one child, which is named...
            assert len(child.children) == 1
            assert child.children[0].name == "SomeThing"  # <... SomeThing
            assert child.children[0].data.type == model.NodeType.SENSOR
            assert child.children[0].data.description == "test"


# TEST COMPLEX INSTANCE STRUCTURE #
# Goal: Check if a list of instantiations works
#
# How: Define a list of different numbers of instances that
# in different ways and check,
# if the result matches every time
#
# TEST COMPLEX INSTANCE STRUCTURE #

# Needed test files
TEST_FILES_COMPLEX = [
    "instance_complex_1.vspec",
    "instance_complex_2.vspec",
    "instance_complex_3.vspec",
    "instance_complex_4.vspec",
]

# test function: read in the files and compare the outcome


def test_complex_structures():
    # helper function to check, if all information are present

    def check_instance_branch(branch: VSSNode, numberOfChildren: int):
        vss_data = branch.get_vss_data()
        assert vss_data.type == model.NodeType.BRANCH
        assert vss_data.description == "High-level vehicle data."

    for tfs in TEST_FILES_COMPLEX:
        print(tfs)
        # load the file
        tree, _ = get_trees(vspec=RESOURCES / tfs, include_dirs=(RESOURCES,))

        # check if root node has 1 child and check first instances
        assert len(tree.children) == 1
        assert tree.children[0].get_fqn() == "Vehicle.Test1"
        check_instance_branch(tree.children[0], 2)

        # go through the instances and check if expected structure maps
        child: VSSNode
        for child in tree.children[0].children:
            # 2nd level: Test1.Test2 - Test1.Test3
            assert child.get_fqn() in [
                "Vehicle.Test1.Test2",
                "Vehicle.Test1.Test3",
            ]
            check_instance_branch(child, 3)
            child_2: VSSNode
            for child_2 in child.children:
                # 2nd level: Test1.Test2.Test4 - Test1.Test3.Test6
                print(child_2.get_fqn())
                assert None is not re.match("^Vehicle.Test1.Test(2|3).Test(4|5|6)", child_2.get_fqn())
                check_instance_branch(child_2, 4)
                child_3: VSSNode
                for child_3 in child_2.children:
                    # 3rd level: Test1.Test2.Test4.Test7 -
                    # Test1.Test3.Test6.Test10
                    assert None is not re.match(
                        "^Vehicle.Test1.Test(2|3).Test(4|5|6).Test(7|8|9|10)",
                        child_3.get_fqn(),
                    )
                    check_instance_branch(child_3, 1)
                    assert 1 == len(child_3.children)
                    child_4 = child_3.children[0]
                    # 4th level: Test1.Test2.Test4.Test7.Test11 -
                    # Test1.Test3.Test6.Test10.Test11
                    assert None is not re.match(
                        "^Vehicle.Test1.Test(2|3).Test(4|5|6).Test(7|8|9|10).Test11",
                        child_4.get_fqn(),
                    )
                    # All instances are expected to have one child
                    assert 1 == len(child_4.children)
                    assert child_4.children[0].name == "SomeThing"
                    assert child_4.children[0].data.type == model.NodeType.SENSOR
                    assert child_4.children[0].data.description == "test"


# TEST EXCLUSION FROM INSTANCE STRUCTURE #
# Goal: Check if exclusion of certain nodes during instantiation work
#
# How: Exclude some nodes from instantiation and check if the
# structure is the expected one
#
# TEST EXCLUSION FROM INSTANCE STRUCTURE #

# Needed test files
TEST_FILES_EXCLUDE = ["instance_exclude_node.vspec"]


def test_exclusion_from_instance():
    for tfs in TEST_FILES_EXCLUDE:
        # load the file
        print(tfs)
        tree, _ = get_trees(vspec=RESOURCES / tfs, include_dirs=(RESOURCES,))
        assert 6 == len(tree.children)
        name_list = []
        child: VSSNode
        for child in tree.children:
            name_list.append(child.get_fqn())
            if "Vehicle.ExcludeSomeThing" == child.get_fqn():
                assert model.NodeType.BRANCH == child.data.type
                assert "ExcludeSomeThing description" == child.data.description
                assert 1 == len(child.children)
                child_2 = child.children[0]
                assert "Vehicle.ExcludeSomeThing.ExcludeSomethingLeaf" == child_2.get_fqn()
                assert "ExcludeSomethingLeaf description" == child_2.data.description
                assert model.NodeType.ACTUATOR == child_2.data.type

            elif "Vehicle.ExcludeNode" == child.get_fqn():
                assert "Vehicle.ExcludeNode" == child.get_fqn()
                assert "ExcludeNode description" == child.data.description
                assert model.NodeType.ATTRIBUTE == child.data.type

        assert "Vehicle.ExcludeSomeThing" in name_list
        assert "Vehicle.ExcludeNode" in name_list


def test_extended_attribute():
    tree, _ = get_trees(vspec=RESOURCES / "instance_extended_attribute.vspec", include_dirs=(RESOURCES,))

    # check if root node has 3 children
    assert len(tree.children) == 3

    # Programmatically change value of dbc/signal for instance in Test2

    child: VSSNode
    for child in tree.children:
        if child.get_fqn() == "Vehicle.Test2":
            assert len(child.children) == 1
            assert child.children[0].name == "SomeThing"  # <... SomeThing
            assert len(child.children[0].data.get_extra_attributes()) == 1
            assert isinstance(getattr(child.children[0].data, "dbc"), dict)
            assert getattr(child.children[0].data, "dbc")["signal"] == "bababa"
            child.children[0].data.dbc["signal"] = "lalala"

    child: VSSNode
    for child in tree.children:
        if child.get_fqn() == "Vehicle.Test2":
            assert child.children[0].data.dbc["signal"] == "lalala"
        else:
            # Make sure that all other instances keeps same name
            assert child.children[0].data.dbc["signal"] == "bababa"



# Copyright (c) 2024 Contributors to COVESA
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License 2.0 which is available at
# https://www.mozilla.org/en-US/MPL/2.0/
#
# SPDX-License-Identifier: MPL-2.0

import subprocess
from pathlib import Path

HERE = Path(__file__).resolve().parent


def test_struct_default(tmp_path):
    vspec = HERE / "struct_default_model_ok.vspec"
    types = HERE / "struct_default_types.vspec"
    cmd = f"vspec export tree -s {vspec} -t {types}"

    # ok
    p = subprocess.run(cmd.split())
    assert p.returncode == 0

    # nok
    log = tmp_path / "log.txt"
    vspec = HERE / "struct_default_model_nok.vspec"
    cmd = f"vspec --log-file {log} export tree -s {vspec} -t {types}"
    p = subprocess.run(cmd.split(), capture_output=True, text=True)
    assert p.returncode != 0
    print(log.read_text())
    assert "invalid 'default' format for datatype 'Types.T'" in log.read_text()



# Copyright (c) 2025 Contributors to COVESA
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License 2.0 which is available at
# https://www.mozilla.org/en-US/MPL/2.0/
#
# SPDX-License-Identifier: MPL-2.0

import logging
from pathlib import Path

import rich_click as click

import vss_tools.cli_options as clo
from vss_tools import log
from vss_tools.lazy_group import LazyGroup


@click.group(context_settings={"auto_envvar_prefix": "vss_tools"}, invoke_without_command=True)
@clo.log_level_opt
@clo.log_file_opt
@click.version_option()
@click.pass_context
def cli(ctx: click.Context, log_level: str, log_file: Path):
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())
    if log_file:
        file_handler = logging.FileHandler(log_file, mode="w")
        file_handler.setFormatter(logging.Formatter("%(asctime)s:%(levelname)s:%(message)s"))
        log.addHandler(file_handler)

    log.setLevel(log_level)


@cli.group(
    cls=LazyGroup,
    lazy_subcommands={
        "apigear": "vss_tools.exporters.apigear:cli",
        "binary": "vss_tools.exporters.binary:cli",
        "csv": "vss_tools.exporters.csv:cli",
        "ddsidl": "vss_tools.exporters.ddsidl:cli",
        "franca": "vss_tools.exporters.franca:cli",
        "plantuml": "vss_tools.exporters.plantuml:cli",
        "id": "vss_tools.exporters.id:cli",
        "json": "vss_tools.exporters.json:cli",
        "jsonschema": "vss_tools.exporters.jsonschema:cli",
        "protobuf": "vss_tools.exporters.protobuf:cli",
        "yaml": "vss_tools.exporters.yaml:cli",
        "tree": "vss_tools.exporters.tree:cli",
        "samm": "vss_tools.exporters.samm:cli",
        "go": "vss_tools.exporters.go:cli",
        "ros2interface": "vss_tools.exporters.ros2interface:cli",
        "s2dm": "vss_tools.exporters.s2dm:cli",
    },
)
@click.pass_context
def export(ctx: click.Context):
    """
    Export a vspec to a chosen format
    """
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())



# Copyright (c) 2024 Contributors to COVESA
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License 2.0 which is available at
# https://www.mozilla.org/en-US/MPL/2.0/
#
# SPDX-License-Identifier: MPL-2.0

from rdflib import URIRef
from vss_tools import log

from ..config import config as cfg


def get_vspec_uri(node_name: str):
    return URIRef(f"{samm_output_namespace}{node_name}")


def get_node_name_from_vspec_uri(node_uri: URIRef):
    return node_uri.replace(samm_output_namespace, "")


def get_unit_uri(unit_name: str):
    return URIRef(f"{samm_base_namespace}:unit:{cfg.SAMM_VERSION}#{unit_name}")


log.debug("VSS to SAMM CONFIG:\n  -- SAMM_TYPE   : %s\n  -- SAMM_VERSION: %s\n", cfg.SAMM_TYPE, cfg.SAMM_VERSION)

# NOTE: samm_base_namespace is more for the ESMF core libraries
samm_prefix = "urn:samm"
samm_base_namespace = f"{samm_prefix}:org.eclipse.esmf.samm"
Namespaces = {
    "samm": f"{samm_base_namespace}:meta-model:{cfg.SAMM_VERSION}#",
    "samm-c": f"{samm_base_namespace}:characteristic:{cfg.SAMM_VERSION}#",
    "samm-e": f"{samm_base_namespace}:entity:{cfg.SAMM_VERSION}#",
    "unit": f"{samm_base_namespace}:unit:{cfg.SAMM_VERSION}#",
}

# Below formatted namespace should look like: urn:samm:com.covesa.vss.spec:5.0.0#
# and is used for the ":" bindings of the converted to TTLs, VSS Aspect models
# that will refer to the user specified output_namespace
samm_output_namespace = f"{samm_prefix}:{cfg.OUTPUT_NAMESPACE}:{cfg.VSPEC_VERSION}#"



# Copyright (c) 2024 Contributors to COVESA
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License 2.0 which is available at
# https://www.mozilla.org/en-US/MPL/2.0/
#
# SPDX-License-Identifier: MPL-2.0


from enum import Enum

from rdflib import URIRef

from ..config import config as cfg
from . import string_helper as str_helper
from .namespaces import samm_base_namespace, samm_output_namespace


class VSSConcepts(Enum):
    EMPTY = ""
    BELONGS_TO = "belongsToVehicleComponent"
    HAS_ATTRIBUTE = "hasStaticVehicleProperty"
    HAS_SIGNAL = "hasDynamicVehicleProperty"
    HAS_COMP_INST = "hasInstance"
    HOLDS_VALUE = "holdsState"
    PART_OF_VEHICLE = "partOfVehicle"
    PART_OF_VEH_COMP = "partOf"
    VEHICLE = "Vehicle"
    VEHICLE_ACT = "ActuatableVehicleProperty"
    VEHICLE_COMP = "VehicleComponent"
    VEHICLE_PROP = "DynamicVehicleProperty"
    VEHICLE_SIGNAL = "ObservableVehicleProperty"
    VEHICLE_STAT = "StaticVehicleProperty"

    def __init__(self, vss_name):
        self.ns = samm_output_namespace
        self.vsso_name = vss_name

    @property
    def uri(self):
        return URIRef(self.uri_string)

    @property
    def uri_string(self):
        return f"{self.ns}{self.value}"


class SammConcepts(Enum):
    ASPECT = "Aspect"
    CHARACTERISTIC = "Characteristic"
    CHARACTERISTIC_RELATION = "characteristic"
    DATA_TYPE = "dataType"
    DESCRIPTION = "description"
    ENTITY = "Entity"
    EVENTS = "events"
    VALUE = "value"
    EXAMPLE_VALUE = "exampleValue"
    NAME = "name"
    OPERATIONS = "operations"
    OPTIONAL = "optional"
    PAYLOAD_NAME = "payloadName"
    PREFERRED_NAME = "preferredName"
    PROPERTIES = "properties"
    PROPERTY = "Property"

    def __init__(self, vss_name):
        self.ns = f"{samm_base_namespace}:meta-model:{cfg.SAMM_VERSION}#"
        self.vsso_name = vss_name

    @property
    def uri(self):
        return URIRef(self.uri_string)

    @property
    def uri_string(self):
        return f"{self.ns}{self.value}"

    @property
    def samm_name(self):
        # Make sure that enum value is lc_first
        return f"samm:{str_helper.str_to_lc_first_camel_case(self.value)}"


class SammCConcepts(Enum):
    BASE_CHARACTERISTICS = "baseCharacteristic"
    BOOLEAN = "Boolean"
    CONSTRAINT = "constraint"
    DEFAULT_VALUE = "defaultValue"
    ENUM = "Enumeration"
    LIST = "List"
    MAX_VALUE = "maxValue"
    MEASUREMENT = "Measurement"
    MIN_VALUE = "minValue"
    QUANTIFIABLE = "Quantifiable"
    RANGE_CONSTRAINT = "RangeConstraint"
    REG_EXP_CONSTRAINT = "RegularExpressionConstraint"
    SINGLE_ENTITY = "SingleEntity"
    STATE = "State"
    TIMESTAMP = "Timestamp"
    TRAIT = "Trait"
    UNIT = "unit"
    VALUES = "values"

    def __init__(self, vss_name):
        self.ns = f"{samm_base_namespace}:characteristic:{cfg.SAMM_VERSION}#"
        self.vsso_name = vss_name

    @property
    def uri(self):
        return URIRef(self.uri_string)

    @property
    def uri_string(self):
        return f"{self.ns}{self.value}"



# Copyright (c) 2021 Contributors to COVESA
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License 2.0 which is available at
# https://www.mozilla.org/en-US/MPL/2.0/
#
# SPDX-License-Identifier: MPL-2.0

# Convert vspec tree to franca


from io import TextIOWrapper
from pathlib import Path

import rich_click as click
from anytree import PreOrderIter

import vss_tools.cli_options as clo
from vss_tools import log
from vss_tools.main import get_trees
from vss_tools.tree import VSSNode


# Write the header line
def print_franca_header(file, version="unknown"):
    file.write(f"""
// Copyright (C) 2022, COVESA
//
// This program is licensed under the terms and conditions of the
// Mozilla Public License, version 2.0.  The full text of the
// Mozilla Public License is at https://www.mozilla.org/MPL/2.0/

const UTF8String VSS_VERSION = "{version}"

struct SignalSpec {{
    UInt32 id
    String name
    String type
    String description
    String datatype
    String unit
    Double min
    Double max
}}

const SignalSpec[] signal_spec = [
""")


# Write the data lines
def print_franca_content(file: TextIOWrapper, root: VSSNode) -> None:
    output = ""
    node: VSSNode
    for node in PreOrderIter(root):
        data = node.get_vss_data()
        if node.parent:
            if output:
                output += ",\n{"
            else:
                output += "{"
            output += f'\tname: "{node.get_fqn()}"'
            output += f',\n\ttype: "{data.type.value}"'
            output += f',\n\tdescription: "{data.description}"'
            datatype = getattr(data, "datatype", None)
            if datatype:
                output += f',\n\tdatatype: "{datatype}"'
            unit = getattr(data, "unit", None)
            if unit:
                output += f',\n\tunit: "{unit}"'
            min = getattr(data, "min", None)
            if min is not None:
                output += f",\n\tmin: {min}"
            max = getattr(data, "max", None)
            if max is not None:
                output += f",\n\tmax: {max}"
            allowed = getattr(data, "allowed", None)
            if allowed:
                output += f",\n\tallowed: {allowed}"
            output += "\n}"
    file.write(output)


@click.command()
@clo.vspec_opt
@clo.output_required_opt
@clo.include_dirs_opt
@clo.extended_attributes_opt
@clo.strict_opt
@clo.types_opt
@clo.aborts_opt
@clo.overlays_opt
@clo.quantities_opt
@clo.units_opt
@clo.strict_exceptions_opt
@click.option("--franca-vss-version", help="Adds franca version info.")
def cli(
    vspec: Path,
    output: Path,
    include_dirs: tuple[Path],
    extended_attributes: tuple[str],
    strict: bool,
    aborts: tuple[str],
    overlays: tuple[Path],
    quantities: tuple[Path],
    units: tuple[Path],
    types: tuple[Path],
    franca_vss_version: str,
    strict_exceptions: Path | None,
):
    """
    Export as Franca.
    """
    log.info("Generating Franca output...")
    tree, _ = get_trees(
        vspec=vspec,
        include_dirs=include_dirs,
        aborts=aborts,
        strict=strict,
        extended_attributes=extended_attributes,
        quantities=quantities,
        types=types,
        units=units,
        overlays=overlays,
        strict_exceptions_file=strict_exceptions,
    )
    with open(output, "w") as f:
        print_franca_header(f, franca_vss_version)
        print_franca_content(f, tree)
        f.write("\n]")



# Copyright (c) 2025 Contributors to COVESA
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License 2.0 which is available at
# https://www.mozilla.org/en-US/MPL/2.0/
#
# SPDX-License-Identifier: MPL-2.0

"""Utilities for modular GraphQL schema export."""

from __future__ import annotations

from pathlib import Path
from typing import Any

from graphql import GraphQLEnumType, GraphQLSchema, is_enum_type, is_object_type, print_type


def analyze_schema_for_flat_domains(schema: GraphQLSchema) -> dict[str, list[str]]:
    """
    Analyze GraphQL schema and create flat domain file structure.

    Each object type gets its own file, named exactly after the type (no transformation).
    Files are organized in domain/ directory with flat structure.
    Instance tags get their own files in instances/ directory.
    Struct types get their own files in structs/ directory.

    Returns:
        Dict mapping file paths to list of type names to include
    """
    domain_files: dict[str, list[str]] = {}

    # Process all object types - put each in its own file with exact type name
    for type_name, graphql_type in schema.type_map.items():
        # Skip built-in types and Query type
        if type_name.startswith("__") or type_name == "Query":
            continue

        if is_object_type(graphql_type):
            # Instance tag types get their own file in instances/
            if "InstanceTag" in type_name:
                instance_file = f"instances/{type_name}.graphql"
                if instance_file not in domain_files:
                    domain_files[instance_file] = []
                domain_files[instance_file].append(type_name)
            # Check if this is a struct type by examining the schema's vspec_comments
            # Structs are identified by having type metadata with vspec_type="STRUCT"
            # Note: We'll need to check this via metadata passed separately
            # For now, use naming convention - types from VehicleDataTypes tree
            elif type_name.startswith("VehicleDataTypes"):
                # Struct types go to structs/ directory
                struct_file = f"structs/{type_name}.graphql"
                domain_files[struct_file] = [type_name]
            else:
                # Regular object types get their own file in domain/
                file_name = f"domain/{type_name}.graphql"
                domain_files[file_name] = [type_name]
        elif is_enum_type(graphql_type):
            # Group enums by category
            if "UnitEnum" in type_name:
                # Unit enums go to other/units.graphql
                enum_file = "other/units.graphql"
                if enum_file not in domain_files:
                    domain_files[enum_file] = []
                domain_files[enum_file].append(type_name)
            elif "InstanceTag" in type_name:
                # Instance tag dimensional enums go with their parent InstanceTag type
                # Extract the base instance tag name (remove _Dimension1, _Dimension2, etc.)
                base_name = type_name
                if "_Dimension" in type_name:
                    base_name = type_name.split("_Dimension")[0]

                instance_file = f"instances/{base_name}.graphql"
                if instance_file not in domain_files:
                    domain_files[instance_file] = []
                domain_files[instance_file].append(type_name)
            elif type_name in ["VspecElementKind", "VspecType"]:
                # Schema meta enums go to directives file (they belong together)
                enum_file = "other/directives.graphql"
                if enum_file not in domain_files:
                    domain_files[enum_file] = []
                domain_files[enum_file].append(type_name)
            else:
                # Domain-specific enums - create their own files in domain/
                enum_file = f"domain/{type_name}.graphql"
                domain_files[enum_file] = [type_name]

    return domain_files


def analyze_schema_for_nested_domains(schema: GraphQLSchema) -> dict[str, list[str]]:
    """
    Analyze GraphQL schema and create nested domain structure based on type relationships.

    Groups related types into domain directories following the hierarchy.
    Instance tags get their own files in instances/ directory.

    Returns:
        Dict mapping file paths to list of type names to include
    """
    # Start from Vehicle type and build hierarchy
    vehicle_type = schema.type_map.get("Vehicle")
    if not vehicle_type or not is_object_type(vehicle_type):
        # Fallback to flat structure if no Vehicle type
        return analyze_schema_for_flat_domains(schema)

    # Group types by their name prefixes (Vehicle_Body -> domain/Vehicle/Body.graphql)
    type_groups: dict[str, list[str]] = {}

    for type_name, graphql_type in schema.type_map.items():
        if type_name.startswith("__") or type_name == "Query":
            continue

        if is_object_type(graphql_type):
            # Instance tag types get their own file in instances/
            if "InstanceTag" in type_name:
                instance_file = f"instances/{type_name}.graphql"
                if instance_file not in type_groups:
                    type_groups[instance_file] = []
                type_groups[instance_file].append(type_name)
            else:
                # Extract domain hierarchy from type name
                if "_" in type_name:
                    # Vehicle_Body_Lights -> domain/Vehicle/Body/Lights.graphql
                    parts = type_name.split("_")
                    if len(parts) >= 2:
                        # Build nested path: domain/Vehicle/Body/Lights.graphql
                        domain_path = "domain/" + "/".join(parts[:-1]) + f"/{parts[-1]}.graphql"
                    else:
                        domain_path = f"domain/{parts[0]}.graphql"
                else:
                    domain_path = f"domain/{type_name}.graphql"

                if domain_path not in type_groups:
                    type_groups[domain_path] = []
                type_groups[domain_path].append(type_name)
        elif is_enum_type(graphql_type):
            # Group enums by category
            if "UnitEnum" in type_name:
                # Unit enums go to other/units.graphql
                enum_file = "other/units.graphql"
                if enum_file not in type_groups:
                    type_groups[enum_file] = []
                type_groups[enum_file].append(type_name)
            elif "InstanceTag" in type_name:
                # Instance tag dimensional enums go with their parent InstanceTag type
                # Extract the base instance tag name (remove _Dimension1, _Dimension2, etc.)
                base_name = type_name
                if "_Dimension" in type_name:
                    base_name = type_name.split("_Dimension")[0]

                instance_file = f"instances/{base_name}.graphql"
                if instance_file not in type_groups:
                    type_groups[instance_file] = []
                type_groups[instance_file].append(type_name)
            elif type_name in ["VspecElementKind", "VspecType"]:
                # Schema meta enums go to directives file (they belong together)
                enum_file = "other/directives.graphql"
                if enum_file not in type_groups:
                    type_groups[enum_file] = []
                type_groups[enum_file].append(type_name)
            else:
                # Domain-specific enums - place in nested domain structure
                if "_" in type_name:
                    # Vehicle_Cabin_DriverPosition_Enum -> domain/vehicle/cabin/DriverPosition_Enum.graphql
                    parts = type_name.split("_")
                    if len(parts) >= 2:
                        # Build nested path
                        enum_path = "domain/" + "/".join(parts[:-1]) + f"/{parts[-1]}.graphql"
                    else:
                        enum_path = f"domain/{parts[0]}.graphql"
                else:
                    enum_path = f"domain/{type_name}.graphql"

                if enum_path not in type_groups:
                    type_groups[enum_path] = []
                type_groups[enum_path].append(type_name)

    return type_groups


def write_domain_files(
    domain_structure: dict[str, list[str]],
    schema: GraphQLSchema,
    output_dir: Path,
    vspec_comments: dict[str, Any],
    directive_processor: Any,
) -> None:
    """
    Write domain-specific GraphQL files.

    Args:
        domain_structure: Mapping of file paths to type names
        schema: GraphQL schema containing the types
        output_dir: Output directory
        vspec_comments: VSS comments for directive processing
        directive_processor: Processor for adding @vspec directives
    """
    for file_path, type_names in domain_structure.items():
        if not type_names:
            continue

        # Create domain file
        domain_file = output_dir / file_path
        domain_file.parent.mkdir(parents=True, exist_ok=True)

        # Special handling for directives file with schema enums
        if file_path == "other/directives.graphql":
            # This file will be handled by write_common_files with schema enums included
            # Skip processing it here to avoid duplication
            continue
        # Special handling for instance files
        if file_path.startswith("instances/"):
            content_parts = ["# Instance tag type and dimensional enums\n"]

            # Separate instance tag types from dimensional enums
            instance_tag_types = []
            dimensional_enums = []

            for type_name in type_names:
                if type_name in schema.type_map:
                    graphql_type = schema.type_map[type_name]
                    if is_object_type(graphql_type) and "InstanceTag" in type_name:
                        instance_tag_types.append(type_name)
                    elif is_enum_type(graphql_type) and "InstanceTag" in type_name:
                        dimensional_enums.append(type_name)

            # Write instance tag types first
            for type_name in instance_tag_types:
                if type_name in schema.type_map:
                    graphql_type = schema.type_map[type_name]
                    type_sdl = print_type(graphql_type)

                    # Add @instanceTag directive to instance tag types
                    if "InstanceTag" in type_name:
                        type_sdl = type_sdl.replace(f"type {type_name}", f"type {type_name} @instanceTag")

                    content_parts.append(type_sdl)

            # Write dimensional enums after instance tag types
            for type_name in dimensional_enums:
                if type_name in schema.type_map:
                    graphql_type = schema.type_map[type_name]
                    type_sdl = print_type(graphql_type)
                    content_parts.append(type_sdl)
        else:
            # Regular domain files
            content_parts = ["# Domain-specific GraphQL types\n"]

            for type_name in type_names:
                if type_name in schema.type_map:
                    graphql_type = schema.type_map[type_name]
                    type_sdl = print_type(graphql_type)

                    # Apply vspec directives if available
                    if directive_processor and hasattr(directive_processor, "add_directives_to_type"):
                        try:
                            enhanced_sdl = directive_processor.add_directives_to_type(
                                type_sdl, type_name, vspec_comments
                            )
                            content_parts.append(enhanced_sdl)
                        except (AttributeError, TypeError):
                            # Fallback to plain SDL if directive processing fails
                            content_parts.append(type_sdl)
                    else:
                        content_parts.append(type_sdl)

        # Write the file
        with open(domain_file, "w") as f:
            f.write("\n\n".join(content_parts))


def write_common_files(
    schema: GraphQLSchema,
    unit_enums_metadata: dict[str, Any],
    allowed_enums_metadata: dict[str, Any],
    vspec_comments: dict[str, Any],
    output_dir: Path,
) -> None:
    """
    Write common GraphQL files (directives, scalars, queries) using GraphQL introspection.

    Args:
        schema: GraphQL schema to extract components from
        unit_enums_metadata: Metadata for unit enums
        allowed_enums_metadata: Metadata for allowed value enums
        vspec_comments: VSS comments for directive processing
        output_dir: Output directory
    """
    from graphql import is_scalar_type, print_type

    from vss_tools.utils.graphql_utils import extract_custom_directives_from_schema

    # Ensure output directory exists
    output_dir.mkdir(parents=True, exist_ok=True)

    # Create other directory for common files
    other_dir = output_dir / "other"
    other_dir.mkdir(parents=True, exist_ok=True)

    # Extract and write directives using GraphQL introspection
    custom_directives = extract_custom_directives_from_schema(schema)

    if custom_directives:
        directives_content = ["# GraphQL directive definitions\n"]

        # For each custom directive, use GraphQL's print function or reconstruct the SDL
        for directive_name, directive_obj in custom_directives.items():
            # Build directive definition string
            args_parts = []
            for arg_name, arg in directive_obj.args.items():
                arg_type_str = str(arg.type)
                if arg.default_value is not None:
                    args_parts.append(f"  {arg_name}: {arg_type_str} = {arg.default_value}")
                else:
                    args_parts.append(f"  {arg_name}: {arg_type_str}")

                # Add description if available
                if arg.description:
                    args_parts[-1] = f'  """{arg.description}"""\n  {args_parts[-1].strip()}'

            # Build locations string
            locations = " | ".join([loc.name for loc in directive_obj.locations])

            # Construct the directive
            directive_def = f"directive @{directive_name}"
            if args_parts:
                directive_def += f"(\n{chr(10).join(args_parts)}\n)"
            directive_def += f" on {locations}"

            directives_content.append(directive_def)

        # Add schema enums that belong with directives (VspecElementKind, VspecType, etc.)
        schema_enums = []
        for type_name, type_def in schema.type_map.items():
            if isinstance(type_def, GraphQLEnumType) and type_name in ["VspecElementKind", "VspecType"]:
                schema_enums.append(print_type(type_def))

        if schema_enums:
            directives_content.extend(schema_enums)

        with open(other_dir / "directives.graphql", "w") as f:
            f.write("\n\n".join(directives_content))

    # Extract custom scalars using GraphQL introspection
    custom_scalars = []
    for type_name, type_def in schema.type_map.items():
        if (
            is_scalar_type(type_def)
            and not type_name.startswith("__")
            and type_name not in ["String", "Int", "Float", "Boolean", "ID"]
        ):
            custom_scalars.append(print_type(type_def))

    if custom_scalars:
        with open(other_dir / "scalars.graphql", "w") as f:
            f.write("# GraphQL scalar type definitions\n\n")
            f.write("\n\n".join(custom_scalars))

    # Extract Query type and schema definition
    query_parts = []
    if "Query" in schema.type_map:
        query_parts.append(print_type(schema.type_map["Query"]))

    # Add schema definition if needed
    if schema.query_type or schema.mutation_type or schema.subscription_type:
        schema_def_parts = ["schema {"]
        if schema.query_type:
            schema_def_parts.append(f"  query: {schema.query_type.name}")
        if schema.mutation_type:
            schema_def_parts.append(f"  mutation: {schema.mutation_type.name}")
        if schema.subscription_type:
            schema_def_parts.append(f"  subscription: {schema.subscription_type.name}")
        schema_def_parts.append("}")
        query_parts.append("\n".join(schema_def_parts))

    if query_parts:
        with open(other_dir / "queries.graphql", "w") as f:
            f.write("# GraphQL query and schema definitions\n\n")
            f.write("\n\n".join(query_parts))



# Copyright (c) 2022 Contributors to COVESA
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License 2.0 which is available at
# https://www.mozilla.org/en-US/MPL/2.0/
#
# SPDX-License-Identifier: MPL-2.0

# Convert vspec tree to binary format
#
# This is the pure Python version of the binary exporter. The
# format has been taken from the C version at
# https://github.com/COVESA/vss-tools/blob/4.2/binary/binarytool.c
# This is a basic "length"/"data" format, where the length is in front
# of each field, and the field order is fixed.
#
# The length fields are mostly 8 bit (i.e. long strings/paths are not
# supported) with the exception of description and allowed (16 bit).
# The order of fields (where each field is composed of
# fieldlength + fielddata) is:
#
# name (vsspath), type, description, datatype, min, max, unit,
# allowed, default, validate
#
# if a field is not present (e.g. min, max, unit, allowed, default, validate),
# the length is 0.

import struct
from pathlib import Path
from typing import BinaryIO

import rich_click as click

import vss_tools.cli_options as clo
from vss_tools import log
from vss_tools.main import get_trees
from vss_tools.tree import VSSNode


def allowedString(allowedList):
    allowedStr = ""
    for elem in allowedList:
        allowedStr += hexAllowedLen(elem) + elem
    return allowedStr


def hexAllowedLen(allowed):
    hexDigit1 = len(allowed) // 16
    hexDigit2 = len(allowed) - hexDigit1 * 16
    return "".join([intToHexChar(hexDigit1), intToHexChar(hexDigit2)])


def intToHexChar(hexInt):
    if hexInt < 10:
        return chr(hexInt + ord("0"))
    else:
        return chr(hexInt - 10 + ord("A"))


# Create a struct containing the length of the string as uint8
#  and the string itself
def create_l8v_string(s: str) -> bytes:
    pack = struct.pack(f"{len(s) + 1}p", s.encode())
    # log.debug(f"create_l8v_string: {s} as {pack.hex()}")
    return pack


# Create a struct containing the length of the string as uint16
#  and the string itself
def create_l16v_string(s: str) -> bytes:
    pack = struct.pack(f"=H{len(s)}s", len(s), s.encode())
    # log.debug(f"create_l16v_string: {s} as {pack.hex()}")
    return pack


def export_node(node: VSSNode, f: BinaryIO):
    data = node.get_vss_data().as_dict()

    f.write(create_l8v_string(node.name))
    f.write(create_l8v_string(data.get("type", "")))
    # Keeping UUID field in output for now (always 0)
    f.write(struct.pack("B", 0))

    f.write(create_l16v_string(data.get("description", "")))
    f.write(create_l8v_string(data.get("datatype", "")))
    f.write(create_l8v_string(str(data.get("min", ""))))
    f.write(create_l8v_string(str(data.get("max", ""))))
    f.write(create_l8v_string(data.get("unit", "")))

    if data.get("allowed") is None:
        f.write(struct.pack("H", 0))
    else:
        f.write(create_l16v_string(allowedString(data.get("allowed", ""))))

    f.write(create_l8v_string(str(data.get("default", ""))))
    f.write(create_l8v_string(str(data.get("validate", ""))))

    f.write(struct.pack("B", len(node.children)))

    for child in node.children:
        export_node(child, f)


@click.command()
@clo.vspec_opt
@clo.output_required_opt
@clo.include_dirs_opt
@clo.extended_attributes_opt
@clo.strict_opt
@clo.aborts_opt
@clo.overlays_opt
@clo.quantities_opt
@clo.units_opt
@clo.types_opt
@clo.types_output_opt
@clo.strict_exceptions_opt
def cli(
    vspec: Path,
    output: Path,
    include_dirs: tuple[Path],
    extended_attributes: tuple[str],
    strict: bool,
    aborts: tuple[str],
    overlays: tuple[Path],
    quantities: tuple[Path],
    units: tuple[Path],
    types: tuple[Path],
    types_output: Path | None,
    strict_exceptions: Path | None,
):
    """
    Export to Binary.
    """

    tree, datatype_tree = get_trees(
        vspec=vspec,
        include_dirs=include_dirs,
        aborts=aborts,
        strict=strict,
        extended_attributes=extended_attributes,
        quantities=quantities,
        units=units,
        types=types,
        overlays=overlays,
        strict_exceptions_file=strict_exceptions,
    )

    log.info("Generating binary output...")
    if datatype_tree:
        if types_output:
            with open(str(types_output), "wb") as f:
                export_node(datatype_tree, f)
            log.info("Binary datatype tree output generated in %s", types_output)
        else:
            log.warning("Ignoring type tree generation as no file name specified")
    with open(str(output), "wb") as f:
        export_node(tree, f)
    log.info("Binary main tree output generated in %s", output)



import nox

PYTHON_VERSIONS = ["3.10", "3.13"]

nox.options.default_venv_backend = "uv"


@nox.session(python=PYTHON_VERSIONS)
def tests(session):
    session.run_install(
        "uv",
        "sync",
        env={"UV_PROJECT_ENVIRONMENT": session.virtualenv.location},
    )
    session.run(
        "pytest",
        "--cov=vss_tools",
        "--cov-report=term-missing",
        "--cov-fail-under=88",
        *session.posargs,
    )



from __future__ import annotations
import yaml
'''
Jinja2 helper for the type mapping.
'''
# Copyright (c) 2021 GENIVI Alliance
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# Generate code that can convert VSS signals to Android Automotive
# vehicle properties.
#

class TypeMap:
    """
    TypeMap helper class for type conversions.
    Reads typemap yaml file from filepath
    Stores locally inside the class
    self.map_tree: map_tree, Parser actual VSS-VHAL map file with possible conversions
    self.typemap: filepath, Reads and parses the filepath yaml file for individual type mapping between VSS/VHAL/CPP.
    self.vss_tree: parsed Vehicle Signal Specification (helper class)
    self.vhal_type: parsed VHAL types map
    """
    def __init__(self,map_tree,filepath,vss_tree,vhal_type):
        self.map_tree=map_tree
        with open(filepath, "r") as f:
            self.typemap = yaml.load(f.read(), Loader=yaml.SafeLoader)
        self.vss_tree=vss_tree
        self.vhal_type=vhal_type

    def __getitem__(self,name):
        """
        Simple item parser. Returns dict that contains the actual mapping.
        name = VSS Type
        """
        return self.typemap[name]

    def vss2cpp(self,key):
        """
        Returns c++ type definition of the VSS item.
        key = VSS Item name
        """
        return self.typemap[self.vss_tree.vsstype(key)]["cpp"]

    def vhal2cpp(self,item):
        """
        Returns c++ type definition of the VHAL item.
        item = VHAL Item name.
        """
        if type(item) is dict:
            return self.typemap[self.vhal_type[item['aospId']]]["cpp"]
        else:
            return self.typemap[self.vhal_type[self.map_tree[item]['aospId']]]["cpp"]

    def vss2vhal(self,key,item=None):
        """
        Returns VHAL type definition of the VSS item.
        key = VSS Item name.
        item = Optional VHAL target type.
        """
        if item is None:
            return self.typemap[self.vss_tree.vsstype(key)]["to"][self.vhal_type[self.map_tree[key]['aospId']]]
        else:
            return self.typemap[self.vss_tree.vsstype(key)]["to"][self.vhal_type[item['aospId']]]

    def vss_from_string(self,key):
        """
        Return c++ type conversion from string to target type.
        key = VSS Item name.
        """
        #typemap[(str(vss_tree[key].data_type).split(".")[-1])]["from"]["string"].replace("_val_","value")
        return self.typemap[str(self.vss_tree[key].data_type).split(".")[-1]]["from"]["string"].replace("_val_","value")

    def vhal_prop(self,item):
        """
        Return VHAL typing for the VehicleProperty.value union.
        """
        if type(item) is dict:
            return "prop.value."+self.typemap[self.vhal_type[item['aospId']]]["vhal"]+"Values"
        else:
            return "prop.value."+self.typemap[self.vhal_type[self.map_tree[item]['aospId']]]["vhal"]+"Values"



# Copyright (c) 2024 Contributors to COVESA
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License 2.0 which is available at
# https://www.mozilla.org/en-US/MPL/2.0/
#
# SPDX-License-Identifier: MPL-2.0

from rdflib import XSD

from .namespaces import get_unit_uri

DataTypes = {
    "uint8": XSD.unsignedByte,
    "int8": XSD.byte,
    "uint16": XSD.unsignedShort,
    "int16": XSD.short,
    "uint32": XSD.unsignedInt,
    "int32": XSD.int,
    "uint64": XSD.unsignedLong,
    "int64": XSD.long,
    "boolean": XSD.boolean,
    "float": XSD.float,
    "double": XSD.double,
    "string": XSD.string,
    "dateTime": XSD.dateTime,
    "dateTimeStamp": XSD.dateTimeStamp,
    "iso8601": XSD.dateTimeStamp,
    "anyURI": XSD.anyURI,
}

DataUnits = {
    "cm3": get_unit_uri("cubicCentimetre"),
    "cm^3": get_unit_uri("cubicCentimetre"),
    "kw": get_unit_uri("kilowatt"),
    "kW": get_unit_uri("kilowatt"),
    "kWh": get_unit_uri("kilowattHour"),
    "l": get_unit_uri("litre"),
    "l/100km": get_unit_uri("litrePerHour"),
    "mm": get_unit_uri("millimetre"),
    "kg": get_unit_uri("kilogram"),
    "inch": get_unit_uri("inch"),
    "A": get_unit_uri("ampere"),
    "Ah": get_unit_uri("ampereHour"),
    "Nm": get_unit_uri("newtonMetre"),
    "N.m": get_unit_uri("newtonMetre"),
    "V": get_unit_uri("volt"),
    "celsius": get_unit_uri("degreeCelsius"),
    "cm/s": get_unit_uri("centimetrePerSecond"),
    "degree": get_unit_uri("degreeUnitOfAngle"),
    "degrees": get_unit_uri("degreeUnitOfAngle"),
    "degrees/s": get_unit_uri("degreePerSecond"),
    "g/s": get_unit_uri("gramPerSecond"),
    "kilometer": get_unit_uri("kilometre"),
    "km": get_unit_uri("kilometre"),
    "km/h": get_unit_uri("kilometrePerHour"),
    "kpa": get_unit_uri("kilopascal"),
    "kPa": get_unit_uri("kilopascal"),
    "l/h": get_unit_uri("litrePerHour"),
    "m": get_unit_uri("metre"),
    "m/s": get_unit_uri("metrePerSecond"),
    "m/s2": get_unit_uri("metrePerSecondSquared"),
    "m/s^2": get_unit_uri("metrePerSecondSquared"),
    "mbar": get_unit_uri("millibar"),
    "min": get_unit_uri("minuteUnitOfTime"),
    "ml": get_unit_uri("millilitre"),
    "pa": get_unit_uri("pascal"),
    "Pa": get_unit_uri("pascal"),
    "percent": get_unit_uri("percent"),
    "percentage": get_unit_uri("percent"),
    "ratio": get_unit_uri("rate"),
    "rpm": get_unit_uri("revolutionsPerMinute"),
    "g/km": get_unit_uri("kilogramPerKilometre"),
    "s": get_unit_uri("secondUnitOfTime"),
    "h": get_unit_uri("secondUnitOfTime"),
    "W": get_unit_uri("watt"),
    "cpm": get_unit_uri("cycle"),
    "bpm": get_unit_uri("cycle"),
    "iso8601": get_unit_uri("secondUnitOfTime"),
    "blank": get_unit_uri("blank"),
}



# Copyright (c) 2022 Contributors to COVESA
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License 2.0 which is available at
# https://www.mozilla.org/en-US/MPL/2.0/
#
# SPDX-License-Identifier: MPL-2.0
import filecmp
import subprocess
from pathlib import Path

HERE = Path(__file__).resolve().parent
TEST_UNITS = HERE / ".." / "test_units.yaml"
TEST_QUANT = HERE / ".." / "test_quantities.yaml"


# Only running json exporter, overlay-functionality should be independent
# of selected exporter
def run_overlay(overlay_prefix, tmp_path):
    spec = HERE / "test.vspec"
    overlay_name = f"overlay_{overlay_prefix}.vspec"
    overlay = HERE / overlay_name
    output = tmp_path / "out.json"
    cmd = f"vspec export json --pretty -u {TEST_UNITS} -q {TEST_QUANT}"
    cmd += f" -e dbc --vspec {spec} -l {overlay} --output {output}"
    subprocess.run(cmd.split(), check=True)
    expected = HERE / f"expected_{overlay_prefix}.json"
    assert filecmp.cmp(output, expected)


def test_explicit_overlay(tmp_path):
    run_overlay("explicit_branches", tmp_path)


def test_implicit_overlay(tmp_path):
    run_overlay("implicit_branches", tmp_path)


def test_no_datatype(tmp_path):
    run_overlay("no_datatype", tmp_path)


def test_no_type(tmp_path):
    run_overlay("no_type", tmp_path)


def test_overlay_error(tmp_path):
    overlay = HERE / "overlay_error.vspec"
    output = tmp_path / "out.json"
    log = tmp_path / "log.txt"
    spec = HERE / "test.vspec"
    cmd = f"vspec --log-file {log} export json --pretty -u {TEST_UNITS}"
    cmd += f" -q {TEST_QUANT} -l {overlay} --vspec {spec} --output {output}"
    process = subprocess.run(cmd.split())
    assert process.returncode != 0
    log_content = log.read_text()
    assert "'A.SignalXXXX' has 1 model error(s)" in log_content
    assert "'type': 'missing'" in log_content
    assert "datatype" in log_content


def test_overlay_branch_error(tmp_path):
    overlay = HERE / "overlay_implicit_branch_no_description.vspec"
    output = tmp_path / "out.json"
    log = tmp_path / "log.txt"
    spec = HERE / "test.vspec"
    cmd = f"vspec --log-file {log} export json --pretty -u {TEST_UNITS}"
    cmd += f" -q {TEST_QUANT} -l {overlay} --vspec {spec} --output {output}"

    process = subprocess.run(cmd.split())
    assert process.returncode != 0
    log_content = log.read_text()
    assert "'A.AB' has 1 model error(s)" in log_content
    assert "'type': 'assertion_error'" in log_content
    assert "description" in log_content



from __future__ import annotations
import yaml
'''
YAML loader for the mapping table.
'''
# Copyright (c) 2021 GENIVI Alliance
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
# Generate code that can convert VSS signals to Android Automotive
# vehicle properties.
#

def load_tree(filepath):
    """
    load_tree(filepath)
    Loads and returns the yaml map file filepath
    Mapping file maps the VSS signals to VHAL signals with possible conversions.
    Breaks the possible elements in conversion formulas to the tree structure for easier utilization in Jinja.
    """
    with open(filepath, "r") as f:
        tree = yaml.load(f.read(), Loader=yaml.SafeLoader)
        # Create input items for the complex yaml formulas - remove arithmetic from the formula for variable evaluation.
        translator=str.maketrans({"/":" ","*":" ","+":" ","(":" ",")":" ","!":" ","&":" ","%":" ","^":" ","|":" ","\t":" "})
        for key,item in tree.items():
            if "translation" in item:
                if "complex" in item["translation"]:
                    formula_elements=item["translation"]["complex"].translate(translator).split()
                    item["translation"]["input"]=[element[1:] for element in formula_elements if element[0]=="$"]
    return tree

if __name__ == "__main__":
    """
    Simple main program to load the example map to the tree (very basic test)
    """
    tree=load_tree("vspec2prop_mapping.yml")
    print(tree)



# Copyright (c) 2024 Contributors to COVESA
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License 2.0 which is available at
# https://www.mozilla.org/en-US/MPL/2.0/
#
# SPDX-License-Identifier: MPL-2.0

from typing import Sequence

from rdflib import Graph, Literal, URIRef
from rdflib.namespace import RDF, XSD
from vss_tools import log
from vss_tools.tree import VSSNode

from . import vss_helper as vss_helper
from .data_types_and_units import DataTypes
from .namespaces import Namespaces, get_node_name_from_vspec_uri, get_vspec_uri
from .samm_concepts import SammCConcepts, SammConcepts, VSSConcepts
from .string_helper import str_camel_case_split, str_to_lc_first_camel_case, str_to_uc_first_camel_case

#
# Builder helper, which provides a set of functions, to set up a TTL Graph,
# build various graph nodes like property, characteristic, entity etc. and add them to this graph,
# which then can be stored in an ESMF Aspect Model Editor loadable TTL file.
#


def __add_node_tuple(graph: Graph, subject_uri: URIRef, predicate_uri: URIRef, object_data: Literal | URIRef) -> bool:
    """
    Helper function to add a Node tuple to specified graph.<br>
    Will validate if the node is present and skip it to avoid duplicates or add is as usual.

    Args:
        graph (Graph): Graph to which node will be added
        subject_uri (URIRef): Subject - URIRef for the node to be created.
        predicate_uri (URIRef): Predicate RDF type for the node to create.
        object_data (Literal | URIRef): Object - contents of the node to create.

    Returns:
        bool: TRUE when node is added to graph and FALSE otherwise.
    """

    # Initialize the node - tuple
    # NOTE: a Tuple elements are: (Subject, Predicate, Object)
    node_tuple = (subject_uri, predicate_uri, object_data)

    # 2: validate if is present
    if graph.__contains__(node_tuple):
        log.debug(
            "A node with below properties already exists."
            "\n  -- subject uri: %s\n  -- predicate: %s\n  -- object data: %s\n",
            subject_uri,
            predicate_uri,
            object_data,
        )

        return False

    # 3: add the tuple to the graph
    else:
        graph.add(node_tuple)

        return True


# Initialize an empty RDF Graph with related bindings and ESMF - AME namespaces
def setup_graph():
    # Create a Graph
    graph = Graph()

    # Bind the namespaces to a prefix for more readable output
    graph.bind(VSSConcepts.EMPTY.value, VSSConcepts.EMPTY.uri)
    graph.bind("xsd", XSD)

    for nsKey in Namespaces:
        graph.bind(nsKey, URIRef(Namespaces[nsKey]))

    return graph


# Build an RDF Tree Node as property for the provided vss_node and add it to the specified graph.
def add_graph_node(graph: Graph, vss_node: VSSNode, is_aspect: bool) -> URIRef:
    node_property_name = vss_helper.get_node_property_name(vss_node)
    node_uri = get_vspec_uri(node_property_name)

    # Initialize the Property node - tuple
    if __add_node_tuple(graph, node_uri, RDF.type, SammConcepts.PROPERTY.uri) is False:
        # Node was already created, just return its URI
        return node_uri

    if is_aspect:
        add_node_aspect(graph, vss_node, node_uri)
    # ELSE: just build a simple property node as usual

    # Preferred name should be white space in front of each upper case letter
    # i.e. more human friendly / readable name
    # EXAMPLE:
    #     node.name     : IsStrongCrossWindDetected
    #     should be like: Is Strong Cross Wind Detected
    __add_node_tuple(
        graph, node_uri, SammConcepts.PREFERRED_NAME.uri, Literal(str_camel_case_split(vss_node.ttl_name), "en")
    )

    log.debug("Created graph node with URI: '%s'.\n", node_uri)

    return node_uri


def add_node_aspect(graph: Graph, vss_node: VSSNode, node_uri: URIRef):
    # Initialize Aspect for current Graph.
    log.debug("Add aspect node for node uri:\t  -- '%s'.", node_uri)

    # NOTE: there can be only 1 Aspect per graph. For Aspect nodes, vss_node.ttl_name is same as vss_node.name
    # Aspect Name must be UC FIRST CAMEL CASE
    aspect_node_uri = get_vspec_uri(str_to_uc_first_camel_case(vss_node.ttl_name))

    if __add_node_tuple(graph, aspect_node_uri, RDF.type, SammConcepts.ASPECT.uri) is True:
        # Aspect added => complete its creation
        __add_node_tuple(
            graph, aspect_node_uri, SammConcepts.PREFERRED_NAME.uri, Literal(str_camel_case_split(vss_node.name), "en")
        )

        # Add the property instance of this Tree node to its Aspect
        __add_node_tuple(
            graph,
            aspect_node_uri,
            SammConcepts.PROPERTIES.uri,
            Literal(f"( {get_property_uri_from_node_uri(node_uri)} )"),
        )

        # Add placeholders for Aspect operations and events
        __add_node_tuple(graph, aspect_node_uri, SammConcepts.OPERATIONS.uri, Literal("()"))
        __add_node_tuple(graph, aspect_node_uri, SammConcepts.EVENTS.uri, Literal("()"))

    # ELSE: do nothing since aspect node has already been created


def add_node_branch_characteristic(graph: Graph, vss_node: VSSNode, node_uri: URIRef):
    # NOTE: constraints are usually on a leaf node
    #       and will result in Node having a Trait with BaseCharacteristic and Constraint nodes.
    node_char_name = get_node_characteristic_name(node_uri, vss_helper.has_constraints(vss_node))
    node_char_uri = get_vspec_uri(node_char_name)

    log.debug("Add branch characteristic: '%s' for VSSNode: '%s'.", node_char_name, vss_node.name)

    if __add_node_tuple(graph, node_char_uri, RDF.type, SammConcepts.CHARACTERISTIC.uri) is False:
        # Characteristic was already added => just return its URI
        return node_char_uri

    # ELSE: complete creation of node characteristic

    __add_node_tuple(graph, node_char_uri, SammConcepts.NAME.uri, Literal(node_char_name))

    # Add description to this node's characteristic
    # NOTE: Keep description under this vss_node's branch - node_char_uri,
    #       because if there is any instances,
    #       each of its instance(s) will point to node_char_uri
    __add_node_tuple(
        graph, node_char_uri, SammConcepts.DESCRIPTION.uri, Literal(vss_helper.get_node_description(vss_node), "en")
    )

    return node_char_uri


def add_node_leaf(graph: Graph, node_uri: URIRef, vss_node: VSSNode):
    log.debug(
        "Add node for VSS Node: '%s'\n  -- of type: '%s'\n  -- with path: '%s'",
        vss_node.name,
        vss_node.data.type,  # type: ignore
        vss_node.get_fqn(),
    )

    has_limits = vss_helper.has_constraints(vss_node)

    node_char_name = get_node_characteristic_name(node_uri, has_limits)
    node_char_uri = get_vspec_uri(node_char_name)

    # Bind current vss_node to its characteristic
    if __add_node_tuple(graph, node_uri, SammConcepts.CHARACTERISTIC_RELATION.uri, node_char_uri) is True:
        # Complete creation of leaf node
        if has_limits:
            node_char_uri = add_node_leaf_constraint(graph, node_char_name, node_char_uri, vss_node)

        # Add description to this node's characteristic
        # NOTE: this could be the general characteristic or trait, in case if vss_node has limits
        __add_node_tuple(
            graph, node_char_uri, SammConcepts.DESCRIPTION.uri, Literal(vss_helper.get_node_description(vss_node), "en")
        )

        # Get RDF and Data types for specified node_characteristic_uri from its related vss_node
        rdf_type = vss_helper.get_node_rdf_type(vss_node)
        data_type = vss_helper.get_data_type(vss_node)

        match rdf_type:
            case SammCConcepts.ENUM | SammCConcepts.STATE:
                # Handle ENUM type nodes
                log.debug("  -- set node ENUM values")

                if vss_node.data.default:  # type: ignore
                    __add_node_tuple(
                        graph,
                        node_char_uri,
                        SammCConcepts.DEFAULT_VALUE.uri,
                        Literal(vss_node.data.default, datatype=data_type),  # type: ignore
                    )

                # Read values for this vss_node's characteristic
                enum_values = None
                if hasattr(vss_node.data, "allowed") and vss_node.data.allowed and type(vss_node.data.allowed) is list:
                    # Add allowed values to this node characteristic
                    enum_values = get_enum_values(vss_node.data.allowed)
                elif hasattr(vss_node.data, "enum") and vss_node.data.enum:
                    # NOTE: From VSS 3.0 the 'enum' attribute has been renamed to `allowed`
                    #       However, we will keep this for backwards compatibility.

                    # Add ENUM values, as usual, to this node characteristic
                    enum_values = get_enum_values(vss_node.data.enum)  # type: ignore

                if enum_values is not None:
                    __add_node_tuple(graph, node_char_uri, SammCConcepts.VALUES.uri, enum_values)

            case SammCConcepts.LIST:
                # Handle LIST type nodes
                log.debug("  -- set node List values")

                # TODO: do we need to add values here?

            case SammCConcepts.MEASUREMENT:
                # Handle MEASUREMENT (unit) type nodes
                log.debug("  -- set node Measurement values")

                if hasattr(vss_node.data, "unit") and vss_node.data.unit:
                    __add_node_tuple(
                        graph, node_char_uri, SammCConcepts.UNIT.uri, vss_helper.get_data_unit_uri(vss_node.data.unit)
                    )

                else:
                    log.warning(
                        "Unit is not set for vss node: '%s' with type: '%s'\n"
                        "Setting node 'RDF.type' from MEASUREMENT to default: CHARACTERISTIC\n",
                        vss_node.name,
                        vss_node.data.datatype,  # type: ignore
                    )

                    rdf_type = SammConcepts.CHARACTERISTIC

                if hasattr(vss_node.data, "default") and vss_node.data.default:
                    __add_node_tuple(
                        graph,
                        node_uri,
                        SammConcepts.EXAMPLE_VALUE.uri,
                        Literal(vss_node.data.default, datatype=data_type),
                    )

            case SammConcepts.CHARACTERISTIC:
                # Handle CHARACTERISTIC type nodes
                log.debug(" -- set regular node values")

                if hasattr(vss_node.data, "default") and vss_node.data.default:
                    if (
                        hasattr(vss_node.data, "unit")
                        and vss_node.data.unit
                        and vss_node.data.unit == "iso8601"
                        and data_type == DataTypes[vss_node.data.unit]
                    ):
                        # Handle date-time based nodes
                        log.debug(" -- set node DATE-TIME values")

                        # NOTE: Some examples are provided in the form: 0000-01-01T00:00Z
                        #       which FAILS ON THE ESMF AME VALIDATION because of the 0000 year.
                        #       Correct format in ESMF AME is:
                        #           yyyy-mm-ddThh:mmZ
                        #       OR
                        #           yyyy-mm-ddThh:mm:ss.milliseconds+hh:mm, where +hh:mm is the timezone
                        #
                        #       Example: '2000-01-01T14:23:00',
                        #                '0001-01-01T00:00:00.00000+00:00',
                        #                '2023-11-27T16:26:05.671Z'
                        if not vss_node.data.default.startswith("0000"):
                            # Add property node exampleValue date time - TIMESTAMP, if is provided and valid
                            __add_node_tuple(
                                graph,
                                node_uri,
                                SammConcepts.EXAMPLE_VALUE.uri,
                                Literal(vss_node.data.default, datatype=data_type),
                            )
                        else:
                            log.warning(
                                "Skipping incorrect date time default value: '%s' for VSS Node: '%s'\n"
                                "CORRECT format + timezone is:\n"
                                "    yyyy-mm-ddThh:mmZ\n"
                                "or\n"
                                "    yyyy-mm-ddThh:mm:ss.milliseconds+hh:mm\n"
                                "where yyyy cannot be just 0000, i.e. 0001 is valid year, but 0000 is not valid.\n",
                                vss_node.data.default,
                                vss_node.name,
                            )

                    else:
                        # Add default ONLY for other than date-time nodes.
                        # DateTime nodes default is handled above.
                        __add_node_tuple(
                            graph,
                            node_uri,
                            SammConcepts.EXAMPLE_VALUE.uri,
                            Literal(vss_node.data.default, datatype=data_type),  # type: ignore
                        )

            case _:
                # DEFAULT
                log.warning(
                    "Could not match Characteristic type: '%s' for vss_node: '%s'\n",
                    rdf_type,
                    vss_node.get_fqn(),
                )

        # Set RDF.type for current leaf node characteristic
        __add_node_tuple(graph, node_char_uri, RDF.type, rdf_type.uri)

        # Set data_type for current vss_node's characteristic
        __add_node_tuple(graph, node_char_uri, SammConcepts.DATA_TYPE.uri, data_type)

    # ELSE: do nothing since leaf node has already been created


def add_node_leaf_constraint(graph: Graph, node_char_name: str, node_char_uri: URIRef, vss_node: VSSNode):
    log.debug("Add leaf-node constraint")

    constraint_name = str_to_uc_first_camel_case(vss_node.ttl_name + "Constraint")
    constraint_node_uri = get_vspec_uri(constraint_name)

    # Default Constraint URI is for Range (min/max) constraints
    constraint_uri = SammCConcepts.RANGE_CONSTRAINT.uri

    if hasattr(vss_node.data, "pattern") and vss_node.data.pattern is not None:
        # Pattern property is used for Regular Expression constraints of STRING based data nodes
        constraint_uri = SammCConcepts.REG_EXP_CONSTRAINT.uri

    __add_node_tuple(graph, constraint_node_uri, RDF.type, constraint_uri)
    __add_node_tuple(graph, constraint_node_uri, SammConcepts.NAME.uri, Literal(constraint_name))

    # Workaround since doubles are serialized as scientific numbers
    data_type = vss_helper.get_data_type(vss_node)
    if data_type == XSD.double:
        data_type = XSD.anyURI

    if vss_node.data.max is not None:  # type: ignore
        __add_node_tuple(
            graph,
            constraint_node_uri,
            SammCConcepts.MAX_VALUE.uri,
            Literal(vss_node.data.max, datatype=data_type),  # type: ignore
        )

    if vss_node.data.min is not None:  # type: ignore
        __add_node_tuple(
            graph,
            constraint_node_uri,
            SammCConcepts.MIN_VALUE.uri,
            Literal(vss_node.data.min, datatype=data_type),  # type: ignore
        )

    if vss_node.data.pattern is not None:  # type: ignore
        __add_node_tuple(
            graph,
            constraint_node_uri,
            SammConcepts.VALUE.uri,
            Literal(vss_node.data.pattern, datatype=data_type),  # type: ignore
        )

        # Set the RegExp value for constraint_node_uri

    base_c_name = str_to_uc_first_camel_case(vss_node.ttl_name + "BaseCharacteristic")
    base_c_uri = get_vspec_uri(base_c_name)

    __add_node_tuple(graph, node_char_uri, SammCConcepts.BASE_CHARACTERISTICS.uri, base_c_uri)
    __add_node_tuple(graph, node_char_uri, RDF.type, SammCConcepts.TRAIT.uri)
    __add_node_tuple(graph, node_char_uri, SammConcepts.NAME.uri, Literal(node_char_name))
    __add_node_tuple(graph, node_char_uri, SammCConcepts.CONSTRAINT.uri, constraint_node_uri)

    return base_c_uri


# Accepts list of:
#  URIRef(s) - URIRef of the property node to be added to specified vss_node_uri
#
# or tuples in the form:
#       ( URIRef, ( "optional", True ), ( "payloadName", "givenName" ) )
# where:
#   - URIRef - is the URIRef of the property node to be added to specified vss_node_uri
#   - other, OPTIONAL, tuples are for property attributes: optional and payloadName
def add_node_properties(vss_nodes_uris: Sequence[URIRef | tuple], graph: Graph, vss_node_uri: URIRef):
    log.debug("Prepare properties from vss nodes URIs:\n%s\n", vss_nodes_uris)

    node_props = ""

    for node_uri in vss_nodes_uris:
        if node_uri:
            property_prefix = " " if node_props else ""

            if type(node_uri) is tuple:
                # Handle node_uri with some additional parameters
                # EXAMPLE tuple:
                #   ( NodeURI, ( "optional", True ), ( "payloadName", "givenName" ) )
                # will result in:
                #   [ samm:property :prop1; samm:optional true; samm:payloadName "givenName"; ]
                property_name = ""
                is_optional = False  # type: ignore
                payload_name = ""

                for entry in node_uri:
                    if type(entry) is URIRef:
                        property_name = get_property_uri_from_node_uri(entry)

                    elif type(entry) is tuple:
                        if entry[0] == "optional":
                            is_optional = entry[1] is True

                        elif entry[0] == "payloadName":
                            # Make sure that payload name starts with lower case
                            payload_name = str_to_lc_first_camel_case(entry[1])

                        else:
                            log.warning(
                                "Node uri tuple: '%s'\n is not in correct format: (attrName, attrValue) "
                                "or is not supported for a property attribute.\n",
                                entry,
                            )

                    else:
                        log.warning(
                            "Node uri tuple: '%s'\n is not in correct format: URIRef OR tuple(attrName, attrValue)\n",
                            entry,
                        )

                if property_name:
                    property_name = f"{SammConcepts.PROPERTY.samm_name} {property_name}"
                    is_optional = f"; {SammConcepts.OPTIONAL.samm_name} true" if is_optional else ""  # type: ignore

                    if payload_name:
                        payload_name = f'; {SammConcepts.PAYLOAD_NAME.samm_name} "{payload_name}"'
                    else:
                        payload_name = ""

                    if is_optional or payload_name:
                        property = f"[ {property_name}{is_optional}{payload_name} ]"
                    else:
                        property = property_name

                    node_props += property_prefix + property

            elif type(node_uri) is URIRef:
                node_props += property_prefix + get_property_uri_from_node_uri(node_uri)

            else:
                log.warning("Not supported type: '%s' for Node URI: '%s'.", type(node_uri), node_uri)

    # Remove trailing white space and set node_props in the form: ( ... )
    node_props = "( {} )".format(node_props.strip())

    log.debug("Add properties URIs to vss node URI: '%s'\n  -- properties:\n%s\n", vss_node_uri, node_props)

    # Add properties to the specified vss_node_uri
    __add_node_tuple(graph, vss_node_uri, SammConcepts.PROPERTIES.uri, Literal(node_props))


def get_property_uri_from_node_uri(vss_node_uri: URIRef):
    return f":{str_to_lc_first_camel_case(vss_node_uri.replace(VSSConcepts.EMPTY.uri, ''))}"


def get_node_characteristic_name(node_uri: URIRef, has_limits: bool):
    # Node characteristic name is based on the node property URI, and should be in the form:
    # NodePropertyNameCharacteristic or NodePropertyNameTrait, in case if the node has some constraints
    node_name = get_node_name_from_vspec_uri(node_uri)
    characteristic_name_suffix = "Trait" if has_limits else "Characteristic"

    return str_to_uc_first_camel_case(node_name + characteristic_name_suffix)


# Helper function to convert a list of strings into TTL formatted ENUM VALUES list
# in the form: ( "value1" "value2" ... "value#" )
# TODO-NOTE: the collection_to_process is provided from VSSNode.allowed, which is currently defined as str.
#            However, ALL allowed definitions in VSS are set as list of strings.
#            For example check VSS::FuelSystem::SupportedFuelTypes::allowed field.
#
#            We might need to further refactor the VssNote.allowed field so it has the correct type: list[str]
#
#            For the moment, we define the collection_to_process as: list[str] | str, so to pass the mypy check.
def get_enum_values(collection_to_process: list[str] | str):
    enum_values = ""

    for value in collection_to_process:
        if type(value) is str:
            enum_values += '"' + value + '" '
        else:
            enum_values += value + " "

    return Literal(f"( {enum_values.strip()} )")


def add_node_instances(graph: Graph, instances_dict_tree: dict, node_char_uri: URIRef):
    instance_char_uri = None

    if instances_dict_tree:
        log.debug("Build nodes from instances dict tree: '%s'", instances_dict_tree["name"])

        # Build Node Instance as characteristic which then will be added to the corresponding node_uri
        node_instance_name = str_to_uc_first_camel_case(instances_dict_tree["name"] + "Instance")
        instance_char_uri, instance_entity_uri = add_node_instance_characteristic_with_entity(graph, node_instance_name)

        instance_entity_properties = []
        skip_siblings_child_nodes = False

        # Populate the node_instance tree graph
        for instance in instances_dict_tree["children"]:
            log.debug(
                "Build ttl node for instance: '%s'\n  -- type: '%s'\n  -- path: '%s'",
                instance["name"],
                instance["type"],
                instance["path"],
            )

            if type(instance) is dict and instance["type"] == "branch":
                log.debug("Build instance path: '%s' as branch node", instance["path"])

                # To make sure we have unique nodes, we use the instance path, similar to VSSNode ttl_name
                ni_path_name = instance["path"].replace(".", "")
                ni_char_name = str_to_uc_first_camel_case(ni_path_name)

                ni_prop_uri = ""
                ni_type_name = ""

                if instance["instance_type"]:
                    # This instance is from a specific type.
                    # For example: instances Row1, Row2, ..., Row# will have an instance_type: Row
                    #
                    # Such instances should share a common characteristic,
                    # i.e. if the instance is DoorRow1, then its characteristic should be: DoorRow,
                    # which later will be shared with all other DoorRow# instances as their shared Characteristic

                    # Instance Type should be prefixed by its parent name to make it more unique, similar to ttl_name
                    ni_type_name = "{}{}".format(instance["parent"]["name"], instance["instance_type"])
                    type_char_name = str_to_uc_first_camel_case(ni_type_name)
                    type_char_uri, type_entity_uri = add_node_instance_characteristic_with_entity(graph, type_char_name)

                    # In this case, the created type_entity will represent a common node for each instance of that type.
                    # Idea is that we can share common type node like characteristic for instances of the same type.
                    ni_entity_uri = type_entity_uri

                    ni_prop_uri = add_node_instance_property(graph, ni_path_name, type_char_uri)

                else:
                    # If the current instance does not have instance_type, just build it as usual
                    ni_char_uri, ni_entity_uri = add_node_instance_characteristic_with_entity(graph, ni_char_name)
                    ni_prop_uri = add_node_instance_property(graph, ni_path_name, ni_char_uri)

                    # Make sure to turn of skip_siblings_child_nodes,
                    # to handle properly children of the current instance
                    skip_siblings_child_nodes = False

                # Add created instance node uri to its entity properties.
                # NOTE: All instance child nodes should be optional
                instance_entity_properties.append((ni_prop_uri, ("optional", True), ("payloadName", instance["name"])))

                if not skip_siblings_child_nodes:
                    # Build instance - entity properties (children) nodes of current instance
                    add_instance_entity_properties(
                        graph, instance["children"], ni_type_name, node_char_uri, ni_entity_uri
                    )

                    if ni_type_name:
                        # Set flag to skip adding properties for instance_typed nodes.
                        # This is to avoid adding of duplicate properties
                        # of instances of same type to their main - type node
                        skip_siblings_child_nodes = True

            elif type(instance) is dict and instance["type"] == "attribute":
                # Attribute instance is like a VSSNode leaf node
                log.debug("Build instance path: '%s' as leaf node.", instance["path"])

                leaf_path_name = instance["path"].replace(".", "")
                leaf_uri = add_node_instance_property(graph, leaf_path_name, node_char_uri)

                instance_entity_properties.append((leaf_uri, ("optional", True), ("payloadName", instance["name"])))

            else:
                log.warning("Instance: '%s' with type: '%s' cannot be processed yet.\n", instance, type(instance))

        add_node_properties(instance_entity_properties, graph, instance_entity_uri)

    return instance_char_uri


# Helper function to support handling of creation of instance nodes for a VSSNode
def add_node_instance_characteristic_with_entity(graph: Graph, node_name: str):
    # NOTE: Node instance characteristic should be of type:
    #       SammCConcepts.SINGLE_ENTITY instead of SammConcepts.CHARACTERISTIC

    # Append the characteristic class (type) to its name uri
    node_char_uri = get_vspec_uri("{}{}".format(node_name, SammCConcepts.SINGLE_ENTITY.vsso_name))

    node_entity_name = str_to_uc_first_camel_case(node_name + "Entity")
    node_entity_uri = get_vspec_uri(node_entity_name)

    if __add_node_tuple(graph, node_char_uri, RDF.type, SammCConcepts.SINGLE_ENTITY.uri) is False:
        # This characteristic is already present, just return its URI and
        return node_char_uri, node_entity_uri

    # ELSE: continue with node creation as usual

    __add_node_tuple(graph, node_char_uri, SammConcepts.NAME.uri, Literal(node_name))
    __add_node_tuple(
        graph, node_char_uri, SammConcepts.PREFERRED_NAME.uri, Literal(str_camel_case_split(node_name), "en")
    )

    # Add the node entity to the graph
    __add_node_tuple(graph, node_entity_uri, RDF.type, SammConcepts.ENTITY.uri)

    # Add the node entity to its characteristic as data type
    __add_node_tuple(graph, node_char_uri, SammConcepts.DATA_TYPE.uri, node_entity_uri)

    return node_char_uri, node_entity_uri


def add_node_instance_property(graph: Graph, instance_name: str, instance_char_uri: URIRef):
    property_name = str_to_lc_first_camel_case(instance_name)
    property_uri = get_vspec_uri(property_name)

    if __add_node_tuple(graph, property_uri, RDF.type, SammConcepts.PROPERTY.uri) is False:
        log.warning("Return: '%s'.\n", property_uri)

        return property_uri

    __add_node_tuple(
        graph, property_uri, SammConcepts.PREFERRED_NAME.uri, Literal(str_camel_case_split(instance_name), "en")
    )

    # Bind this instance node to the provided instance_char_uri
    __add_node_tuple(graph, property_uri, SammConcepts.CHARACTERISTIC_RELATION.uri, instance_char_uri)

    return property_uri


def add_instance_entity_properties(
    graph: Graph, instance_children: list[dict], instance_type: str, characteristic_uri: URIRef, entity_uri: URIRef
):
    if instance_children and len(instance_children) > 0:
        entity_properties = []

        for child_instance in instance_children:
            child_path_name = ""
            if instance_type:
                # Use parent instance type + name for a child name
                child_path_name = "{}{}".format(instance_type, child_instance["name"])
            else:
                # Use usual child path for its name, similar to VSSNode ttl_name
                child_path_name = child_instance["path"].replace(".", "")

            # Add node instance as property node with characteristic of its main - parent VSSNode
            # NOTE: at this point each instance property node will be linked to the VSS defined node (characteristic)
            # FOR EXAMPLE: Door.Row1.DriverSide instance will have a DoorCharacteristic node as defined in VSS.
            child_uri = add_node_instance_property(graph, child_path_name, characteristic_uri)

            # Instance child properties should be optional and with payloadName without its parent prefix.
            # For more details check function: add_node_properties
            entity_properties.append((child_uri, ("optional", True), ("payloadName", child_instance["name"])))

        # Add child properties nodes to this instance's entity node
        add_node_properties(entity_properties, graph, entity_uri)



# Copyright (c) 2021 Contributors to COVESA
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License 2.0 which is available at
# https://www.mozilla.org/en-US/MPL/2.0/
#
# SPDX-License-Identifier: MPL-2.0

# Convert vspec file to proto
#

import sys
from io import TextIOWrapper
from pathlib import Path

import rich_click as click
from anytree import findall

import vss_tools.cli_options as clo
from vss_tools import log
from vss_tools.main import get_trees
from vss_tools.model import (
    VSSDataBranch,
    VSSDataDatatype,
    VSSDataStruct,
)
from vss_tools.tree import VSSNode

PATH_DELIMITER = "."
DIR_DELIMITER = "/"

mapped = {
    "uint16": "uint32",
    "uint8": "uint32",
    "int8": "int32",
    "int16": "int32",
    "boolean": "bool",
}


def init_package_file(path: Path, package_name: str):
    with open(path, "w") as f:
        log.info(f"Initializing {path}, package {package_name}")
        f.write('syntax = "proto3";\n\n')
        f.write(f"package {package_name};\n\n")


def traverse_data_type_tree(tree: VSSNode, static_uid: bool, add_optional: bool, out_dir: Path):
    """
    All structs in a branch are written to a single .proto file.
    The file's base name is same as the branch's name
    The fully qualified path of the type becomes the package.

    Example A.B.C.MyType becomes:
    "package A.B.C;
    message MyType{}"

    The files are organized by package.
    With the above example, the file generated is
    A/B/C/C.proto
    """

    node: VSSNode
    for node in findall(tree, filter_=lambda n: isinstance(n.data, VSSDataStruct)):
        # A/B/C/MyType
        struct_path = Path(node.get_fqn("/"))

        # <out-dir>/A/B/C
        package_path = out_dir / struct_path.parent
        package_path.mkdir(parents=True, exist_ok=True)

        # <out-dir>/A/B/C/C.proto
        out_file = package_path / f"{package_path.name}.proto"
        if not out_file.exists():
            init_package_file(out_file, ".".join(struct_path.parent.parts))

        with open(out_file, "a") as fd:
            imports = []
            for c_node in findall(node, filter_=lambda n: isinstance(n.data, VSSDataDatatype)):
                datatype = c_node.data.datatype
                if "." not in datatype:
                    continue
                c_struct_path = Path(datatype.replace(".", "/"))
                if c_struct_path.parent != struct_path.parent:
                    imports.append(f"{c_struct_path.parent}/{c_struct_path.parent.name}.proto")

            write_imports(fd, imports)

            fd.write(f"message {struct_path.name} {{" + "\n")
            print_messages(node.children, fd, static_uid, add_optional)
            fd.write("}\n\n")
            log.info(f"Wrote {struct_path.name} to {out_file}")


def traverse_signal_tree(tree: VSSNode, fd: TextIOWrapper, static_uid: bool, add_optional: bool):
    fd.write('syntax = "proto3";\n\n')

    imports = []
    for node in findall(tree, filter_=lambda n: isinstance(n.data, VSSDataDatatype)):
        datatype = node.data.datatype
        if "." not in datatype:
            continue
        struct_path = Path(node.data.datatype.replace(".", "/"))  # type: ignore
        imports.append(f"{struct_path.parent}/{struct_path.parent.name}.proto")
    write_imports(fd, imports)

    # write proto messages to file
    for node in findall(tree, filter_=lambda node: isinstance(node.data, VSSDataBranch)):
        fd.write(f"message {node.get_fqn('')} {{" + "\n")
        print_messages(node.children, fd, static_uid, add_optional)
        fd.write("}\n\n")


def write_imports(fd: TextIOWrapper, imports: list[str]):
    imports = list(set(imports))
    imports.sort()
    for stmt in imports:
        fd.write(f'import "{stmt}";\n')
    fd.write("\n")


def print_messages(nodes: tuple[VSSNode], fd: TextIOWrapper, static_uid: bool, add_optional: bool):
    usedKeys: dict[int, str] = {}
    for i, node in enumerate(nodes, 1):
        if isinstance(node.data, VSSDataDatatype):
            dt_val = node.data.datatype
            data_type = mapped.get(dt_val.strip("[]"), dt_val.strip("[]"))
            if dt_val.endswith("[]"):
                data_type = "repeated " + data_type
            elif add_optional:
                data_type = "optional " + data_type
        else:
            data_type = node.get_fqn("")
            if add_optional:
                data_type = "optional " + data_type
        if static_uid:
            if "staticUID" not in node.data.get_extra_attributes():
                log.fatal(
                    (
                        f"Aborting because {node.get_fqn()} does not have the staticUID attribute. "
                        f"When using the option --static-uid each node must have the attribute staticUID."
                    )
                )
                sys.exit(-1)
            fieldNumber = int(int(getattr(node.data, "staticUID"), 0) / 8)
            if fieldNumber < 20000 and fieldNumber >= 19000:
                log.fatal("""Aborting because field number {fieldNumber} for signal {node.name} is in
                reservered range between 19000 and 20000. Consider changing the signal to alter the staticUID.""")
                sys.exit(-1)
            if fieldNumber in usedKeys:
                log.fatal(
                    (
                        f"Aborting, due to collision for fieldNumber {fieldNumber}. "
                        f"It is used by {node.get_fqn()} and {usedKeys[fieldNumber]}. "
                        "Consider changing the signals to alter the staticUID."
                    )
                )
                fd.truncate(0)
                sys.exit(-1)
            else:
                usedKeys[fieldNumber] = node.get_fqn()
        else:
            fieldNumber = i
        fd.write(f"  {data_type} {node.name} = {fieldNumber};" + "\n")


@click.command()
@clo.vspec_opt
@clo.output_required_opt
@clo.include_dirs_opt
@clo.extended_attributes_opt
@clo.strict_opt
@clo.aborts_opt
@clo.overlays_opt
@clo.quantities_opt
@clo.units_opt
@clo.types_opt
@clo.strict_exceptions_opt
@click.option(
    "--types-out-dir",
    help="Output directory for generated protos.",
    type=click.Path(file_okay=False, writable=True),
)
@click.option(
    "--static-uid",
    is_flag=True,
    help="Expect staticUID attribute in the vspec input and use it as field number",
)
@click.option("--add-optional", is_flag=True, help="Set each field to 'optional'")
def cli(
    vspec: Path,
    output: Path,
    include_dirs: tuple[Path],
    extended_attributes: tuple[str],
    strict: bool,
    aborts: tuple[str],
    overlays: tuple[Path],
    quantities: tuple[Path],
    units: tuple[Path],
    types: tuple[Path],
    types_out_dir: Path | None,
    static_uid: bool,
    add_optional: bool,
    strict_exceptions: Path | None,
):
    """
    Export as protobuf.
    """
    log.info("Generating protobuf output...")
    tree, datatype_tree = get_trees(
        vspec=vspec,
        include_dirs=include_dirs,
        aborts=aborts,
        strict=strict,
        extended_attributes=extended_attributes,
        quantities=quantities,
        units=units,
        types=types,
        overlays=overlays,
        strict_exceptions_file=strict_exceptions,
    )
    if datatype_tree:
        if not types_out_dir:
            types_out_dir = Path.cwd()
            log.warning(f"No output directory given. Writing to: {types_out_dir.absolute()}")
        traverse_data_type_tree(datatype_tree, static_uid, add_optional, types_out_dir)

    with open(output, "w") as f:
        log.info(f"Writing to: {output}")
        traverse_signal_tree(tree, f, static_uid, add_optional)



# Copyright (c) 2023 Contributors to COVESA
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License 2.0 which is available at
# https://www.mozilla.org/en-US/MPL/2.0/
#
# SPDX-License-Identifier: MPL-2.0
import subprocess
from pathlib import Path

import pytest

HERE = Path(__file__).resolve().parent
DATA = HERE / "data"


@pytest.mark.parametrize(
    "spec_id, quantity_id, unit_id, expected_out",
    [
        (1, 1, 1, "Invalid datatype: 'numerics'"),
        (1, 1, 2, "Invalid quantity: 'workie'"),
        (1, 1, 3, "Field required"),
        (2, 1, 4, "'arraysize' set on a non array datatype"),
        (3, 1, 4, "'foo' is not of type 'uint8[]'"),
        (4, 1, 4, "'min/max' and 'allowed' cannot be used together"),
        (5, 1, 4, "default value '4' is not in 'allowed' list"),
        (6, 1, 4, "'string' is not allowed for unit 'kWh'"),
        (7, 1, 4, "Forbidden extra attribute (core attribute): 'Vehicle':'datatype'"),
        (8, 1, 4, "'default' array size does not match 'arraysize'"),
    ],
)
def test_consistency(spec_id: int, quantity_id: int, unit_id: int, expected_out: str, tmp_path: Path):
    spec = DATA / f"s_{spec_id}.yaml"
    quantity = DATA / f"q_{quantity_id}.yaml"
    unit = DATA / f"u_{unit_id}.yaml"
    log = tmp_path / "log.txt"
    cmd = f"vspec --log-file {log} export tree -s {spec} -u {unit} -q {quantity}"
    process = subprocess.run(cmd.split(), capture_output=True, text=True)
    assert process.returncode != 0
    assert expected_out in log.read_text()



# Copyright (c) 2024 Contributors to COVESA
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License 2.0 which is available at
# https://www.mozilla.org/en-US/MPL/2.0/
#
# SPDX-License-Identifier: MPL-2.0

"""
GraphQL scalar types and datatype mappings for VSS exporters.

This module provides reusable GraphQL scalar type definitions and mappings
that can be used across different VSS exporters. It centralizes the
custom scalar types needed for VSS data types.
"""

from graphql import (
    GraphQLBoolean,
    GraphQLFloat,
    GraphQLInt,
    GraphQLScalarType,
    GraphQLString,
)

from vss_tools.datatypes import Datatypes

# Custom scalar types for VSS data types
VSS_SCALARS = {
    "Int8": GraphQLScalarType(name="Int8"),
    "UInt8": GraphQLScalarType(name="UInt8"),
    "Int16": GraphQLScalarType(name="Int16"),
    "UInt16": GraphQLScalarType(name="UInt16"),
    "UInt32": GraphQLScalarType(name="UInt32"),
    "Int64": GraphQLScalarType(name="Int64"),
    "UInt64": GraphQLScalarType(name="UInt64"),
}

# Mapping from VSS datatypes to GraphQL types
VSS_DATATYPE_MAP = {
    Datatypes.INT8[0]: VSS_SCALARS["Int8"],
    Datatypes.UINT8[0]: VSS_SCALARS["UInt8"],
    Datatypes.INT16[0]: VSS_SCALARS["Int16"],
    Datatypes.UINT16[0]: VSS_SCALARS["UInt16"],
    Datatypes.INT32[0]: GraphQLInt,
    Datatypes.UINT32[0]: VSS_SCALARS["UInt32"],
    Datatypes.INT64[0]: VSS_SCALARS["Int64"],
    Datatypes.UINT64[0]: VSS_SCALARS["UInt64"],
    Datatypes.FLOAT[0]: GraphQLFloat,
    Datatypes.DOUBLE[0]: GraphQLFloat,
    Datatypes.BOOLEAN[0]: GraphQLBoolean,
    Datatypes.STRING[0]: GraphQLString,
}


def get_vss_scalar_types() -> list[GraphQLScalarType]:
    """Get list of all VSS custom scalar types for schema inclusion."""
    return list(VSS_SCALARS.values())



# Copyright (c) 2023 Contributors to COVESA
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License 2.0 which is available at
# https://www.mozilla.org/en-US/MPL/2.0/
#
# SPDX-License-Identifier: MPL-2.0
import filecmp
import subprocess
from pathlib import Path
from typing import Optional

HERE = Path(__file__).resolve().parent
TEST_UNITS = HERE / ".." / "test_units.yaml"


def run_unit(
    tmp_path,
    vspec_file,
    unit_argument,
    expected_file,
    quantity_argument="",
    grep_present: bool = True,
    grep_string: Optional[str] = None,
    fails: bool = False,
):
    out = tmp_path / "out.json"
    log = tmp_path / "log.txt"
    cmd = f"vspec --log-file {log} export json --pretty --vspec {vspec_file}"
    cmd += f" {unit_argument} {quantity_argument} --output {out}"
    process = subprocess.run(cmd.split(), capture_output=True, text=True, cwd=HERE)
    print(process.stdout)
    if fails:
        assert process.returncode != 0
    else:
        assert process.returncode == 0
        assert filecmp.cmp(out, HERE / expected_file)

    if grep_string and grep_present:
        assert grep_string in log.read_text() or grep_string in process.stderr


def test_default_unit_no_quantity_warning(tmp_path):
    """
    If no quantity file is found it shall only inform about that
    """
    run_unit(
        tmp_path,
        "test.vspec",
        "-u ../test_units.yaml",
        None,
        "",
        True,
        "No 'quantity' files defined.",
        True,
    )
    run_unit(
        tmp_path,
        "test.vspec",
        "-u ../test_units.yaml",
        None,
        "",
        False,
        "Invalid quantity: 'length'",
        True,
    )


def test_default_unit_quantities_missing(tmp_path):
    """
    If quantity file found it shall inform about missing quantities (once for each)
    """
    run_unit(
        tmp_path,
        "test.vspec",
        "-u ../test_units.yaml",
        None,
        "-q test_quantities.yaml",
        False,
        "Invalid quantity: 'length'",
        True,
    )


def test_default_unit_quantities_ok(tmp_path):
    run_unit(
        tmp_path,
        "test.vspec",
        "-u ../test_units.yaml",
        "expected.json",
        "-q ../test_quantities.yaml",
        False,
        "No quantities defined",
    )
    run_unit(
        tmp_path,
        "test.vspec",
        "-u ../test_units.yaml",
        "expected.json",
        "-q ../test_quantities.yaml",
        False,
        "Quantity length used by unit km has not been defined",
    )
    run_unit(
        tmp_path,
        "test.vspec",
        "-u ../test_units.yaml",
        "expected.json",
        "-q ../test_quantities.yaml",
        False,
        "Quantity temperature used by unit celsius has not been defined",
    )
